/***************************************************************************
                          fshpeerinfo.c  -  description
                             ---------------------
    begin                : 20 februari 2003
    copyright            : 
    author               : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *  This is a basic sample to demonstrate how to insert or delete or       *
 *  update data in the table using parameters                              *
 *                                                                         *
 ***************************************************************************/


#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <dlfcn.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdarg.h>
#include <ctype.h>
#include <malloc.h>
#include <fcntl.h>
#include <math.h>
#include <mysql.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <sys/ioctl.h>
#include <net/if.h>
#ifdef SOLARIS
#include <sys/sockio.h>
#endif

#include "fshpeerinfo.h" /* MyODBC 3.51 sample utility header */

#define USE_OPENSSL 1
#define DEBUG_MALLOC 0
#define HOSTKEY_LEN 2048
#define EXTRA_CHECKS YES
#define MAX_TEMP_HOSTS 32
#define MAX_INTERFACES 16
#define HELOEXPIRES 1440
#define MAX_HELO_EXPIRES (60 * 60 * 24 * 10)
#define MAX_BUFFER_SIZE 65536
#define INTERFACES "eth0"
#define IP_ADDRESS "62.131.97.197"
#define BASICLIMITING YES
#define MAXNETDOWNBPSTOTAL 50000
#define MAXNETUPBPSTOTAL 50000
#define MAXNETLOAD 80
#define MAXCPULOAD 2000
#define DISKQUOTA 1024
#define p2p_PROTO_HELO 0
#define DSO "udp tcp"
#define DSOEXT  ".so"
#define HOSTDIR "/usr/local/fsh/gnunet/var/data/hosts/"
#define LOGFILE "/usr/local/fsh/gnunet/var/log.txt"
#define HOSTKEYFILE "/usr/local/fsh/gnunet/var/.hostkey"

/* do not turn this on unless you know what you
   are doing, you'll get a ton of output... */
#define DEBUG_LOCKING 0

#define OK 1
#define SYSERR -1

#define YES 1
#define NO 0

#define UDP_PROTOCOL_NUMBER 17
#define MAX_PROC_LINE 5000

#define HELO_Message_size(helo) ((sizeof(HELO_Message) + ntohs((helo)->senderAddressSize)))
#define PRIP(ip) (ip)>>24, (ip)>>16 & 255, (ip)>>8 & 255, (ip) & 255
#define STRDUP(a) xstrdup(a,__FILE__,__LINE__)
/* where to read network interface information from
   under Linux */
#define PROC_NET_DEV "/proc/net/dev"
/* amount of kbs gnunetd tries to keep free at all times */
#define MIN_FREE 1024

typedef struct {
  unsigned long long last_in;
  unsigned long long last_out;
} NetworkStats;

#define LOG_NOTHING    0
#define LOG_FATAL      1
#define LOG_ERROR      2
#define LOG_FAILURE    3
#define LOG_WARNING    4
#define LOG_MESSAGE    5
#define LOG_INFO       6
#define LOG_DEBUG      7
#define LOG_CRON       8
#define LOG_EVERYTHING 9

/* use IFLOG(LOG_XXX, statement(s)) for statements
   that should only be executed if we are at the
   right loglevel */
#define IFLOG(a,b) {if (getLogLevel() >= a) {b;} }

#define ANY_PROTOCOL_NUMBER 0

#define DEBUG_CRON 0

/* use these constants to specify timeings */
#define cronMILLIS 1
#define cronSECONDS 1000 * cronMILLIS
#define cronMINUTES 60 * cronSECONDS
#define cronHOURS 60 * cronMINUTES
#define cronDAYS 24 * cronHOURS
#define cronWEEKS 7 * cronDAYS
#define cronMONTHS 30 * cronDAYS
#define cronYEARS 365 * cronDAYS

/** the initial size of the cron-job table **/
#define INIT_CRON_JOBS 16

/* change this value to artificially speed up all
   GNUnet cron timers by this factor. E.g. with 10,
   a cron-job scheduled after 1 minute in the code
   will occur after 6 seconds. This is useful for
   testing bugs that would otherwise occur only after
   a long time. 

   For releases, this value should always be 1 */
#define SPEED_UP 1

/** number of cron units (ms) in a second **/
#define CRON_UNIT_TO_SECONDS (1000 / SPEED_UP)

/** number of ns [usec] in a cron-unit (1000) **/
#define NANOSEC_TO_CRON_UNIT (1000 * SPEED_UP)

#ifdef USE_OPENSSL
#include <openssl/ripemd.h>
#include <openssl/rsa.h>
#include <openssl/evp.h>
#include <openssl/bn.h>
#define HOSTKEY RSA *

#else

#include <gcrypt.h>
#define HOSTKEY GCRY_SEXP

#endif

#ifndef PTHREAD_MUTEX_NORMAL
#ifdef PTHREAD_MUTEX_TIMED_NP
#define PTHREAD_MUTEX_NORMAL PTHREAD_MUTEX_TIMED_NP
#else
#define PTHREAD_MUTEX_NORMAL NULL
#endif
#endif

#if DEBUG_CRON > 1
#define HAVE_PRINT_CRON_TAB 1
#else
#define HAVE_PRINT_CRON_TAB 0
#endif

typedef struct Mutex {
  pthread_mutex_t pthreadMutex;
} Mutex;

typedef struct Semaphore {
  int         v;
  Mutex mutex;
  pthread_cond_t cond;
} Semaphore;

/* ************** Semaphore operations ************* */
Semaphore * new_semaphore (int value);
void free_semaphore(Semaphore * s);
int semaphore_down (Semaphore * s);
int semaphore_up (Semaphore * s);

/* ************** MUTEX OPERATIONS ***************** */
#if DEBUG_LOCKING
#define MUTEX_CREATE(a) do { \
  fprintf(stderr, \
          "Creating mutex %x at line %d in file %s\n", \
          (int) a, __LINE__, __FILE__); \
  create_mutex_(a); \
}\
while(0)
#define MUTEX_CREATE_RECURSIVE(a) do { \
  fprintf(stderr, \
          "Creating recursive mutex %x at line %d in file %s\n", \
          (int) a, __LINE__, __FILE__); \
  create_recursive_mutex_(a); \
}\
while(0)
#define MUTEX_DESTROY(a) do { \
  fprintf(stderr, \
          "Destroying mutex %x at line %d in file %s\n", \
          (int) a, __LINE__, __FILE__); \
  destroy_mutex_(a); \
}\
while(0)
#define MUTEX_LOCK(a) do { \
  fprintf(stderr, \
          "Aquireing lock %x at line %d in file %s\n", \
          (int)a, __LINE__, __FILE__); \
  mutex_lock_(a); \
}\
while (0)
#define MUTEX_UNLOCK(a) do { \
  fprintf(stderr, \
         "Releasing lock %x at line %d in file %s\n", \
	(int)a,__LINE__, __FILE__); \
  mutex_unlock_(a); \
}\
while (0)
#else
#define MUTEX_LOCK(a) mutex_lock_(a)
#define MUTEX_UNLOCK(a) mutex_unlock_(a)
#define MUTEX_CREATE(a) create_mutex_(a)
#define MUTEX_CREATE_RECURSIVE(a) create_recursive_mutex_(a)
#define MUTEX_DESTROY(a) destroy_mutex_(a)
#endif

/* while we must define these, use the macros instead! */
void create_mutex_(Mutex * mutex);
void create_recursive_mutex_(Mutex * mutex);
void create_fast_mutex_(Mutex * mutex);
void destroy_mutex_(Mutex * mutex);
void mutex_lock_(Mutex * mutex);
void mutex_unlock_(Mutex * mutex); 

/**
 * This prototype is somehow missing in various Linux pthread
 * include files. But we need it and it seems to be available
 * on all pthread-systems so far. Odd.
 **/
extern int pthread_mutexattr_setkind_np(pthread_mutexattr_t *attr, int kind);

typedef unsigned long long cron_t; 

typedef void (*CronJob)(void *);

/**
 * The Delta-list for the cron jobs.
 **/
typedef struct {
  /** The start-time for this event (in milliseconds). */
  cron_t delta;
  /** The method to call at that point. */
  CronJob method;
  /** for cron-jobs: when this should be repeated
      automatically, 0 if this was a once-only job */
  unsigned int deltaRepeat;
  /** The index of the next entry in the delta list
      after this one */
  int next;
  /** data ptr (argument to the method) **/
  void * data;
} DeltaEntry;

/* *************** thread handling ************* */


/**
 * Print information about the peer.
 * Currently prints only the HostIdentity
 * and the IP. Could of course do more
 * (e.g. resolve via DNS or print trust).
 **/ 

/**
 * A 160-bit hashcode
 **/
typedef struct {
  int a;
  int b;
  int c;
  int d;
  int e;
} HashCode160;
 
typedef void (*DirectoryEntryCallback)(char * filename, char * dirName, void * data);

typedef struct {
  HashCode160 hashPubKey;
} HostIdentity;

typedef struct {
  /**
   * size of this MESSAGE_PART (network byte order) 
   **/
  unsigned short size;

  /**
   * type of the request, XX_p2p_PROTO_XXX (network byte order) 
   **/
  unsigned short requestType; 

  /**
   * this is followed by a requestType specific data block,
   * consisting of size bytes. 
   **/
  char data[0];
} p2p_HEADER;

/**
 * Length of RSA encrypted data (2048 bit)
 **/
#define RSA_ENC_LEN 256

/**
 * Length of an RSA KEY (d,e,len), 2048 bit (=256 octests) key d, 2 byte e
 **/
#define RSA_KEY_LEN 258

typedef struct {
  unsigned char sig[RSA_ENC_LEN]; 
} Signature;

typedef struct {
  unsigned short len; /*  in big-endian, must be RSA_KEY_LEN+2 */
  unsigned short sizen;  /*  in big-endian! */ 
  unsigned char key[RSA_KEY_LEN];
  unsigned short padding; /* padding (must be 0) */
} PublicKey;

typedef struct {
  p2p_HEADER header;


  /**
   * The signature 
   **/
  Signature signature;

  /**
   * The public key 
   **/
  PublicKey publicKey; 

  /**
   * Whose identity follows? No, this is NOT a duplicate
   * as a node may send us the identity of ANOTHER node! 
   **/
  HostIdentity senderIdentity; 

  /**
   * time this address expires  (network byte order) 
   **/ 
  time_t expirationTime;

  /**
   * size of the sender address 
   **/
  unsigned short senderAddressSize;

  /**
   * protocol supported by the node (only one protocol
   * can be advertised by the same HELO) 
   * Examples are UDP, TCP, etc. This field is
   * in network byte order 
   **/
  unsigned short protocol;

  /**
   * advertised MTU for sending (replies can have a different
   * MTU!) 
   **/
  unsigned int MTU;

  /**
   * address of the node in a protocol specific format 
   **/ 
  char senderAddress[0]; 

} HELO_Message;

/**
 * Hash2Hex: filename, consisting of 32 bytes [0-9A-Z] plus
 * null-termination.
 **/
typedef struct {
  unsigned char data[sizeof(HashCode160)*2+1];
} HexName;

struct TempStorage_ {
  HexName hex;
  HELO_Message * helo;
  int result;
};

typedef struct {
  unsigned short len; /*  in big-endian! */
  unsigned short sizen;/*  in big-endian! */
  unsigned short sizee;/*  in big-endian! */
  unsigned short sized;/*  in big-endian! */
  unsigned short sizep;/*  in big-endian! */
  unsigned short sizeq;/*  in big-endian! */
  unsigned short sizedmp1;/*  in big-endian! */
  unsigned short sizedmq1;/*  in big-endian! */
  unsigned char key[0];
} HostKeyEncoded;

typedef struct {
  HostIdentity identity;
  /** how long is this host blacklisted? **/
  cron_t until;
  /** what would be the next increment for blacklisting? (in seconds) **/
  unsigned int delta;
  /** for which protocol is this host known? */
  unsigned short protocol;
} HostEntry;

/**
 * Opaque handle for a session representation on the transport
 * layer side 
 **/
typedef struct {
  unsigned short ttype;
  void * internal;
} TSession;

/**
 * Type of an iterator over all hosts.
 * @param identity the identity of the host
 * @param protocol the available protocol
 * @param data the data-argument passed to forEachHost
 **/
typedef void (*HostIterator)(const HostIdentity * identity, const unsigned short protocol, void * data);

/**
 * This header file contains a list of the methods that every
 * transport layer implementation must provide. The basic idea is that
 * gnunetd calls "inittransport_XXX" on every transport-api, passing a
 * struct with gnunetd core services to the transport api, and getting
 * a struct with services provided by the transport api back (or null
 * on error). The return value of init is of type TransportAPI.
 *
 * Example:
 *
 * TransportAPI * inittransport_XXX(CoreTransportAPI * api) {
 *   if (api->version != 0)
 *     return NULL;
 *   // ...
 *   return myApi;
 * }
 *
 * The type of inittransport_XXX is TransportMainMethod.
 **/
typedef struct {

  /**
   * This field is used by the core internally;
   * the transport should never do ANYTHING
   * with it.
   **/
  void * libHandle;

  /**
   * The name of the transport, set by the
   * core. Read only for the service itself!
   **/ 
  char * transName;

  /**
   * The number of the protocol that is supported by this transport
   * API (i.e. 6 tcp, 17 udp, 80 http, 25 smtp, etc.)
   **/
  unsigned short protocolNumber;

  /**
   * The MTU for the protocol (e.g. 1472 for UDP).
   * Can be up to 65535 for stream-oriented transport
   * protocols)
   **/
  unsigned short mtu;

  /**
   * How costly is this transport protocol (compared to the other
   * transports, UDP and TCP are scaled to be both 100). The cost is
   * used by GNUnet to select the most preferable mode of
   * transportation.
   **/
  unsigned int cost;

  /**
   * Verify that a HELO-Message is correct (a node
   * is potentially reachable at that address). Core
   * will only play ping pong after this verification passed.
   * @param helo the HELO message to verify
   *        (the signature/crc have been verified before)
   * @return OK if the helo is well-formed
   **/
  int (*verifyHelo)(HELO_Message * helo);
  
  /**
   * Create a HELO-Message for the current node. The HELO is
   * created without signature, timestamp, senderIdentity
   * or publicKey. The GNUnet core will sign the message 
   * and add these other fields. The callee is only
   * responsible for filling in the protocol number, 
   * senderAddressSize and the senderAddress itself.
   *
   * @param helo address where to store the pointer to the HELO
   *        message
   * @return OK on success, SYSERR on error (e.g. send-only
   *  transports return SYSERR here)
   **/
  int (*createHELO)(HELO_Message ** helo);

  /**
   * Establish a connection to a remote node.
   *
   * @param helo the HELO-Message for the target node
   * @param tsession the session handle that is to be set
   * @return OK on success, SYSERR if the operation failed
   **/
  int (*connect)(HELO_Message * helo,
		 TSession ** tsession);

  /**
   * Send a message to the specified remote node.
   * @param tsession an opaque session handle (e.g. a socket
   *        or the HELO_message from connect)
   * @param msg the message
   * @param size the size of the message, <= mtu
   * @return SYSERR on error, OK on success; after any error,
   *         the caller must call "disconnect" and not continue
   *         using the session afterwards (useful if the other
   *         side closed the connection).
   **/
  int (*send)(TSession * tsession,
	      const void * msg,
	      const unsigned int size,
	      int isEncrypted,
	      const int crc);

  /**
   * A (core) Session is to be associated with a transport session. The
   * transport service may want to know in order to call back on the
   * core if the connection is being closed. Associate can also be
   * called to test if it would be possible to associate the session
   * later, in this case, call disconnect afterwards. This can be used
   * to test if the connection must be closed by the core or if the core
   * can assume that it is going to be self-managed (if associate
   * returns OK and session was NULL, the transport layer is responsible
   * for eventually freeing resources associated with the tesession). If
   * session is not NULL, the core takes responsbility for eventually
   * calling disconnect.
   * 
   * @param tsession the session handle passed along
   *   from the call to receive that was made by the transport
   *   layer
   * @return OK if the session could be associated,
   *         SYSERR if not.
   **/
  int (*associate)(TSession * tsession);

  /**
   * Disconnect from a remote node. A session can be closed
   * by either the transport layer calling "closeSession" on
   * the core API or by the core API calling "disconnect"
   * on the transport API. Neither closeSession nor
   * disconnect should call the other method. Due to 
   * potentially concurrent actions (both sides close the
   * connection simultaneously), either API must tolerate
   * being called from the other side.
   *
   * @param tsession the session that is to be closed
   * @return OK on success, SYSERR if the operation failed
   **/
  int (*disconnect)(TSession * tsession);  

  /**
   * Start the server process to receive inbound traffic.
   * @return OK on success, SYSERR if the operation failed
   **/
  int (*startTransportServer)(void);

  /**
   * Shutdown the server process (stop receiving inbound
   * traffic). Maybe restarted later!
   **/
  int (*stopTransportServer)(void);

  /**
   * Reload the configuration. Should never fail (keep old
   * configuration on error, syslog errors!)
   **/
  void (*reloadConfiguration)(void);

  /**
   * Convert transport address to human readable string.
   **/
  char * (*addressToString)(HELO_Message * helo);

} TransportAPI;

/**
 * Type of the per-transport callback method.
 **/ 
typedef void (*TransportCallback)(TransportAPI * tapi, void * data);

/**
 * Type of a struct passed to receive.
 **/
typedef struct {
  /**
   * The session associated with the message
   * on the transport layer side. Maybe passed to "associate"
   * in order to send replies on a bi-directional pipe (if
   * possible).
   **/
  TSession * tsession;

  /**
   * The identity of the sender node
   **/
  HostIdentity sender;

  /**
   * The message itself. The GNUnet core will call 'xfree' once
   * processing of msg is complete. Note that msg can point to
   * multiple p2p_headers.
   **/
  p2p_HEADER * msg;

  /**
   * The size of the message
   **/
  unsigned int size;
  
  /**
   * YES if the message was encrypted, NO otherwise
   **/
  int isEncrypted;

  /**
   * The checksum of the message (over size bytes from msg)
   **/
  int crc;
} MessagePack;

/**
 * This header file contains a draft for the gnunetd
 * core API. This API is used by the transport layer
 * for communication with the GNUnet core.
 * 
 * A pointer to an instance of this struct is passed
 * to the init method of each Transport API.
 **/
typedef struct {

  /**
   * The version of the CORE API. For now, always "0".
   **/
  unsigned int version;

  /**
   * The identity of the local node.
   **/
  HostIdentity * myIdentity;

  /**
   * Data was received (potentially encrypted), make
   * the core process it.
   *
   * @param mp the message, freed by the callee once processed!
   **/
  void (*receive)(MessagePack * mp);

} CoreAPIForTransport;

typedef TransportAPI * (*TransportMainMethod)(CoreAPIForTransport *);

typedef int (*CommandLineParser)(int argc, char * argv[]);

typedef void (*DriveFullCallback)(long long * spaceLeft);
typedef struct {
	struct in_addr addr;
} IPaddr;

/**
 * Struct describing a network in CIDR notation.
 **/
typedef struct {
  IPaddr network;
  IPaddr netmask;
} CIDRNetwork;

typedef struct UserConfStruct {
  char * section;
  char * option;
  char * stringValue;
  unsigned int intValue;
  struct UserConfStruct * next;
} UserConf;

struct CFG_ENTRIES {
    int  ent_count;
    char **ent_names;
    char **ent_values;
    int  **ent_seen;
};

struct CFG_SECTIONS {
    int                 sec_count;
    char                **sec_names;
    struct CFG_ENTRIES  **sec_entries;
};

/**
 * Host-Address in a UDP network.
 **/
typedef struct {
  /**
   * claimed IP of the sender, network byte order 
   **/  
  IPaddr senderIP;

  /**
   * claimed port of the sender, network byte order 
   **/
  unsigned short senderPort; 

  /**
   * reserved (set to 0 for signature verification) 
   **/
  unsigned short reserved; 

} HostAddress;

struct CFG_SECTIONS * c = NULL;

unsigned char * encoding__ = "0123456789ABCDEF";
unsigned char hostsdirectory[255]; 

FILE * logfile = NULL;
int loglevel__ = LOG_WARNING;
int timestamping = YES;
Mutex logMutex;
Mutex tapis_lock;
Mutex deltaListLock_;
Mutex statLock;
Mutex configLock;
Mutex statLock;
Mutex statusMutex;
SQLHENV    henv;
SQLHDBC    hdbc; 
SQLHSTMT   hstmt;
SQLINTEGER narg;
UserConf * userconfig = NULL;
int parseConfigInit = NO;

/**
 * The current (allocated) size of knownHosts
 **/
int max_;

/**
 * The number of actual entries in knownHosts
 **/
int count_;

/**
 * A lock for accessing knownHosts
 **/
Mutex lock_;

/**
 * The list of known hosts.
 **/
HostEntry * hosts_;

/**
 * Directory where the HELOs are stored in (data/hosts)
 **/
char * networkIdDirectory;

/**
 * The list of temporarily known hosts
 **/
HELO_Message * tempHosts[MAX_TEMP_HOSTS];

/**
 * tempHosts is a ringbuffer, this is the current
 * index into it.
 **/
int tempHostsNextSlot;

int tapis_count;
/**
 * Our current IP address.
 **/
IPaddr myAddress;

TransportAPI ** tapis;
HostIdentity myIdentity;
HOSTKEY hostkey;
PublicKey * publicKey;
CoreAPIForTransport transportCore;
/**
 * The delta-list of waiting tasks.
 **/
DeltaEntry * deltaList_;

/**
 * The current size of the DeltaList.
 **/
unsigned int deltaListSize_;

/**
 * The first empty slot in the delta-list.
 **/
int firstFree_;

/**
 * The first empty slot in the delta-list.
 **/
int firstUsed_;

/**
 * The cron thread.
 **/
int cron_shutdown = YES;
pid_t cronPID = 0;
Semaphore * cron_signal = NULL;

/* ***************** global statistics ***************** */

/**
 * When did the module start? 
 **/
cron_t startTime;

/**
 * How many values do we keep statistics for? 
 **/
int statCounters;

/**
 * What are these values (value) 
 **/
unsigned long long * values=NULL;

/**
 * A description for each of the values 
 **/
char ** descriptions=NULL;

char ** interfacePtrs; /* Pointer to the name of each interface, has numInterfaces entries */
int numInterfaces = 0; /* how many interfaces do we have? */

/* configuration */
int maxNetDownBPS;
int maxNetUpBPS;
int maxNetLoad; 		   /* in percent of maxNetXXXBPS */
int maxCPULoad; 		   /* in percent of 1 CPU */
int lastNetResultUp = -1;   /* the max upstream load we saw last time */
int lastNetResultDown = -1; /* the max dnstream load we saw last time */
cron_t lastnettime   =  0;  /* when did we check last time? */
int useBasicMethod = YES;   /* how to measure traffic */

/* tracking */
char ** tracked_files = NULL;
int tracked_file_count = 0;
NetworkStats * last_net_results; /* has numInterfaces entries */
long long spaceLeft; /* in KB */
long long spaceUsed; /* in KB */
DriveFullCallback * driveFullCallbacks;
int dfc_count;

/* quota */
unsigned int diskQuota; /* in MB! */
int stat_handle_kb_used;
int stat_handle_kb_remaining;
int stat_handle_network_load;
int stat_handle_cpu_load;

/**
 * Flag to ensure that we do not start a
 * second job while we are already
 * cleaning up...
 **/
Semaphore * signalSemaphore = NULL;

NetworkStats globalTrafficBetweenProc;

char * expandDollar(char * section, char * orig);

/**
 * This method does not really belong here, but
 * where else to put it...
 * @return a random value in the interval [0,i[. 
 **/
int randomi(int i) {
  return (int) (((float)i)*rand()/(RAND_MAX+1.0));
}

/**
 * Get the cost of a message in for the given transport mechanism.
 **/
unsigned int transportGetCost(int ttype) 
{
  TransportAPI * tapi;
  
  if (ttype >= tapis_count) return SYSERR; /* -1 = INFTY */
  tapi = tapis[ttype];
  if (tapi == NULL) return SYSERR; /* -1 = INFTY */
  return tapi->cost;
}


/**
 * Convert a textual description of a loglevel into an int.
 **/
int getLogLevel() 
{
  return loglevel__;
}

/**
 * Print the current time to logfile without linefeed
 **/
static void printTime() 
{
  if ( (timestamping == YES) && (logfile !=NULL) ) 
    {
      char timebuf[64];
      time_t timetmp;
      struct tm * tmptr;
 
      time(&timetmp);
      tmptr = localtime(&timetmp);
      strftime(timebuf, 64, "%b %e %H:%M:%S ", tmptr);
      fputs(timebuf, logfile);
    }
}

void breakpoint() 
{
  if (logfile != NULL) 
    {
      printTime();
      fprintf(logfile, "__BREAK__");
      fflush(logfile);
    } 
  else
    fprintf(stderr, "__BREAK__");
}

/**
 * errexit - log an error message and exit.
 *
 * @param format the string describing the error message
 **/
void errexit(const char *format, ...) 
{
  va_list args;

  /* NO locking here, we're supposed to die,
     and we don't want to take chances on that... */
  va_start(args, format);
  if (logfile != NULL) 
    {
      printTime();
      vfprintf(logfile, format, args);
      fflush(logfile);
    } 
  else
    vfprintf(stderr, format, args);
  va_end(args);
  breakpoint();
  exit(1);
}
 
/**
 * Convert a textual description of a loglevel into an int.
 **/
static int getLoglevel(char * log) 
{
  if (log == NULL)errexit("LOGLEVEL specified is NULL, that's not ok.\n");
  if (0 == strcmp(log, "NOTHING"))return LOG_NOTHING;
  if (0 == strcmp(log, "FATAL"))return LOG_FATAL;
  if (0 == strcmp(log, "ERROR"))return LOG_ERROR;
  if (0 == strcmp(log, "FAILURE"))return LOG_FAILURE;
  if (0 == strcmp(log, "WARNING"))return LOG_WARNING;
  if (0 == strcmp(log, "MESSAGE"))return LOG_MESSAGE;
  if (0 == strcmp(log, "INFO"))return LOG_INFO;
  if (0 == strcmp(log, "DEBUG"))return LOG_DEBUG;
  if (0 == strcmp(log, "CRON"))return LOG_CRON;
  if (0 == strcmp(log, "EVERYTHING"))return LOG_EVERYTHING;
  errexit("invalid loglevel specified: %s (did you use all-caps?)\n", log);
  return 42; /* can not happen */
}

/**
 * Re-read the loggig configuration.
 * Call on SIGHUP if the configuration file has changed.
 **/
void resetLogging() {
  /* char * loglevelname;
  char * logfilename; */

  MUTEX_LOCK(&logMutex);
  
  /* loglevelname
    = getConfigurationString("GNUNETD",
			     "LOGLEVEL");
  loglevel__ 
    = getLoglevel(loglevelname);
  xfree(loglevelname,
	"freeing string of loglevel");

  timestamping
    = testConfigurationString("GNUNETD",
			      "TIMESTAMPS",
			      "YES"); 
  logfilename
    = getConfigurationString("GNUNETD",
			     "LOGFILE"); */

  /* if (logfilename != NULL) */
  if (LOGFILE != NULL) 
    {
      /* char * fn; */

      /* fn = expandFileName(logfilename); */

      /* logfile = fopen(fn, "a+"); */

      logfile = fopen(LOGFILE, "a+");

      /* xfree(fn, "resetLogging - logfilename");
      xfree(logfilename, "log file name - before expansion"); */
    } 
  else
    logfile = stderr;

  MUTEX_UNLOCK(&logMutex);
}

/**
 * Log a debug message
 * @param format the string describing the error message
 **/
void LOG(int minLogLevel,
	 const char *format, ...) 
{
  va_list	args;  

  if (loglevel__ < minLogLevel)
    return;

  MUTEX_LOCK(&logMutex);
  va_start(args, format);
  if (logfile != NULL) 
    {
      printTime();
      vfprintf(logfile, format, args);
      fflush(logfile);
    } 
  else
    vfprintf(stderr, format, args);
  va_end(args);

  MUTEX_UNLOCK(&logMutex);
}

/**
 * Shutdown the logging module.
 **/
void doneLogging() 
{
  if ( (logfile != NULL) &&(logfile != stderr) ) fclose(logfile);
  logfile = NULL;
  loglevel__ = 0;
  MUTEX_DESTROY(&logMutex);
}

/**
 * Allocate memory. Checks the return value, aborts if no more
 * memory is available.
 **/
void * xmalloc(int size, const char * site) 
{
  void * result;

  result = malloc(size);
#if DEBUG_MALLOC > 2
  printf("xmalloc: allocating %d bytes (%x) at site %s\n", size, (int) result, site);
  fflush(stdout);
#endif  
#if DEBUG_MALLOC == 2
  printf("%x malloc %s\n", (int)result, site);
  fflush(stdout);
#endif
  if (result == NULL) 
    {
      errexit("ERROR: out of memory (%d at %s)!\n", size, site);
    }
  memset(result, 0, size); /* client code should not rely on this, though... */
  return result;
}

/**
 * Free memory. Merely a wrapper for the case that we
 * want to keep track of allocations.
 **/
void xfree(void * ptr, 
	   const char * site) 
{
#if DEBUG_MALLOC
  if (ptr == NULL) errexit("xfree called with NULL ptr at %s\n", site);
#if DEBUG_MALLOC > 2
  printf("xfree: %s (%x)\n", site, (int) ptr);
  fflush(stdout);
#endif
#if DEBUG_MALLOC == 2
  printf("%x free %s\n", (int) ptr, site);
  fflush(stdout);
#endif
#endif
  free(ptr);
}

/* -------------------------------------------------------------------- */
/* The following functions replace standard library functions in that   */
/* they exit on any error returned from the system calls.  Saves us     */
/* from having to check each and every call above.                      */
/* -------------------------------------------------------------------- */

static int tw_pthread_cond_wait(pthread_cond_t * c, 
				pthread_mutex_t * m) 
{
  int return_value;
  
  if ((return_value = pthread_cond_wait (c, m)) != 0) errexit("FATAL: pthread_cond_wait returned %d (%s)\n", return_value, strerror(return_value));  	    
  return 0;
}

static int tw_pthread_cond_signal (pthread_cond_t * c) 
{
  int return_value;

  if ((return_value = pthread_cond_signal (c)) != 0) errexit("FATAL: pthread_cond_signal returned %d (%s)\n", return_value, strerror(return_value));  
  return 0;
}

/* ********************* public methods ******************* */

void create_mutex_(Mutex * mutex) 
{
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
  #if USE_CHECKING_MUTEX
  #ifdef LINUX
    pthread_mutexattr_setkind_np(&attr, PTHREAD_MUTEX_ERRORCHECK_NP);  
  #endif
  #else
    pthread_mutexattr_setkind_np(&attr, (int)PTHREAD_MUTEX_NORMAL);
  #endif
    if ( pthread_mutex_init(&mutex->pthreadMutex, &attr) != 0) errexit("FATAL: creation of mutex failed (%s)", strerror(errno));
}

void create_recursive_mutex_(Mutex * mutex) 
{
  pthread_mutexattr_t attr;

  pthread_mutexattr_init(&attr);
  #ifdef LINUX
    pthread_mutexattr_setkind_np(&attr, PTHREAD_MUTEX_RECURSIVE_NP);
  #elif SOMEBSD
    pthread_mutexattr_setkind_np(&attr, PTHREAD_MUTEX_RECURSIVE); 
  #elif SOLARIS
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
  #endif
    if ( pthread_mutex_init(&mutex->pthreadMutex, &attr) != 0) errexit("FATAL: creation of recursive mutex failed");
}

void destroy_mutex_(Mutex * mutex) 
{
  pthread_mutex_destroy(&mutex->pthreadMutex);
}

void mutex_lock_(Mutex * mutex) 
{
  int return_value;
  
  if ((return_value = pthread_mutex_lock (&mutex->pthreadMutex)) != 0) errexit("FATAL: pthread_mutex_lock returned %d (%s)\n", return_value, strerror(return_value));  
}

void mutex_unlock_(Mutex * mutex) 
{
  int return_value;

  if ((return_value = pthread_mutex_unlock(&mutex->pthreadMutex)) != 0) errexit("FATAL: pthread_mutex_unlock returned %d (%s)\n", return_value, strerror(return_value));  
}

/**
 * function must be called prior to semaphore use -- handles
 * setup and initialization.  semaphore destroy (below) should
 * be called when the semaphore is no longer needed.
 **/
Semaphore * new_semaphore (int value) 
{
  Semaphore * s = xmalloc(sizeof(Semaphore), "new_semaphore: semaphore");
  s->v = value;
  MUTEX_CREATE(&(s->mutex));

  if (pthread_cond_init (&(s->cond), NULL) != 0) errexit("FATAL: pthread_cont_init failed\n");
  return s;
}

void free_semaphore(Semaphore * s) 
{
  MUTEX_DESTROY(&(s->mutex));
  pthread_cond_destroy(&(s->cond));
  xfree(s, "free_semaphore");
}

/**
 * function increments the semaphore and signals any threads that
 * are blocked waiting a change in the semaphore.
 **/
int semaphore_up (Semaphore * s) 
{
  int value_after_op;

  MUTEX_LOCK(&(s->mutex));  
  (s->v)++;
  value_after_op = s->v;  
  MUTEX_UNLOCK(&(s->mutex));
  tw_pthread_cond_signal (&(s->cond));  
  return (value_after_op);
}

/**
 * function decrements the semaphore and blocks if the semaphore is
 * <= 0 until another thread signals a change.
 **/
int semaphore_down (Semaphore * s) 
{
  int value_after_op;
  
  MUTEX_LOCK(&(s->mutex));
  while (s->v <= 0) 
    {
      tw_pthread_cond_wait (&(s->cond), &(s->mutex.pthreadMutex));
    }
  (s->v)--;
  value_after_op = s->v;
  MUTEX_UNLOCK(&(s->mutex));  
  return (value_after_op);
}

/**
 * Create a thread. Use this method instead of pthread_create since
 * BSD may only give a 1k stack otherwise.
 *
 * @param stackSize the size of the stack of the thread in bytes. 
 *        Note that if the stack overflows, some OSes (seen under BSD) 
 *        will just segfault and gdb will give a messed-up stacktrace.
 * @return see pthread_create 
 **/
int PTHREAD_CREATE(pthread_t * handle,
		   void * (*main)(void *),
		   void * arg,
		   size_t stackSize) 
{
  pthread_attr_t stack_size_custom_attr;

  pthread_attr_init(&stack_size_custom_attr);
  pthread_attr_setstacksize(&stack_size_custom_attr, stackSize);
  return pthread_create(handle, &stack_size_custom_attr, main, arg);			    
}

/**
 * Free if not NULL.
 **/
void freeNonNull(void * ptr,
		 const char * site) 
{
  if (ptr != NULL) xfree(ptr, site);
}

/**
 * Dup a string.
 **/
char * xstrdup(const char * str,
	       const char * filename,
	       const int linenumber) 
{
  char * res;

  res = strdup(str);
  #if DEBUG_MALLOC == 2
    printf("%x strdup %s:%d\n", (int) res, filename, linenumber);
    fflush(stdout);
  #endif
  return res;
}

/**
 * Hash block of given size.
 * @param block the data to hash, length is given as a second argument
 * @param size the length of the data to hash
 * @param ret pointer to where to write the hashcode
 **/
void hash(void * block,
	  int size,
	  HashCode160 * ret) 
{
  #ifdef USE_OPENSSL 
	RIPEMD160(block, size, (unsigned char*) ret);
  #else
    gcry_md_hash_buffer(GCRY_MD_RMD160, (char*) ret, block, size);
#endif
}

/**
 * Convert (hash) block to hex (= filename)
 * @param block the sequence to convert
 * @param result where to store thestring (\0-terminated), hex-encoding
 **/
void hash2hex(const HashCode160 * block,
	      HexName * result) 
{
  int i;
  int j;
  unsigned char c;
  unsigned char clow;

  if ((block == NULL) || (result == NULL)) errexit("hash2hex called with block or result NULL!\n");

  result->data[sizeof(HashCode160)*2]=0;
  j=0;
  for (i=0;i<sizeof(HashCode160);i++) 
     {
       c = ((unsigned char *)block)[i]; 
       clow = c & 15; /* get lower nibble */
       result->data[j++] = encoding__[clow];
       clow = c >> 4; /* get higher nibble */
       result->data[j++] = encoding__[clow];
     }  
}

/**
 * Scan a directory for files. The name of the directory
 * must be expanded first (!).
 * @param dirName the name of the directory
 * @param callback the method to call for each file,
 *        can be NULL, in that case, we only count
 * @param data argument to pass to callback
 * @return the number of files found, -1 on error
 **/
int scanDirectory(char * dirName,
		  DirectoryEntryCallback callback,
		  void * data) 
{
  DIR * dinfo;
  struct dirent *finfo;
  struct stat istat;
  int count = 0;
 
  if (dirName == NULL) return -1; 
  if (0 != stat(dirName, &istat)) 
    {
      LOG(LOG_WARNING, "WARNING: Could not stat %s (%s)\n", dirName, strerror(errno));
      return -1;
    }
  if (!S_ISDIR(istat.st_mode)) 
    {
      LOG(LOG_ERROR, "ERROR: scanDirectory must be invoked on a directory (%s)!\n", dirName);
      return -1;
    }
  errno = 0;
  dinfo = opendir(dirName);
  if ((errno == EACCES) || (dinfo == NULL)) 
    {
      LOG(LOG_WARNING, "WARNING: scanDirectory: %s (%s)\n", strerror(errno), (char*)dirName);
      return -1;
    }
  while ((finfo = readdir(dinfo)) != NULL) 
    {
      if (finfo->d_name[0] == '.') continue;
      if (callback != NULL)
        callback(finfo->d_name, dirName, data);
      count++;
    }
  closedir(dinfo);
  return count;
}

/**
 * Is this transport mechanism available (for sending)?
 * @return YES or NO
 **/
int isTransportAvailable(unsigned short ttype) 
{
  /* printf("tapis_count=%d ttype=%d!\n", tapis_count, ttype); */

  if (ttype >= tapis_count) return NO;
  if (NULL == tapis[ttype]) return NO;
  return YES;
}

/**
 * Get the filename under which we would store the HELO_Message
 * for the given host and protocol.
 * @return filename of the form DIRECTORY/HOSTID.PROTOCOL
 **/
static char * getHostFileName(const HostIdentity * id,
			      const unsigned short protocol) 
{
  HexName fil;
  char * fn;

  hash2hex(&id->hashPubKey, &fil);
  fn = xmalloc(strlen(networkIdDirectory) + sizeof(HexName) + 1 + 5 + 1, /* '.' (1) + %u (5) + '\0' (1) */"getHostFileName: filename");
  sprintf(fn, "%s%s.%u", networkIdDirectory,(char*) &fil, protocol);
  return fn;
}

/**
 * Read the contents of a binary file into a buffer.
 * @param fileName the name of the file, not freed,
 *        must already be expanded!
 * @param len the maximum number of bytes to read
 * @param result the buffer to write the result to
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readFile(char * fileName,
	     int  len,
	     void * result) {
  /* open file, must exist, open read only */
  int handle;  
  int size;

  if ((fileName == NULL) || (result == NULL))
    return -1;
  handle = open(fileName,O_RDONLY,S_IRUSR);
  if (handle < 0)
    return -1; 
  size = read(handle, result, len);
  close(handle);
  return size;  
}

/**
 * Compare two hashcodes.
 * @return 1 if they are equal, 0 if not.
 **/
int equalsHashCode160(const HashCode160 * a, 
			     const HashCode160 * b) {
  return (0 == memcmp(a,b,sizeof(HashCode160)));
}

/**
 * Check if 2 hosts are the same.
 * @return YES if they are equal, otherwise NO
 **/
int hostIdentityEquals(const HostIdentity * first, 
		       const HostIdentity * second) {
  if ( (first == NULL) || 
       (second == NULL) )
    return NO;
  return equalsHashCode160(&first->hashPubKey,
			   &second->hashPubKey);
}

/**
 * Check if the filename matches the identity that we are searching
 * for. If yes, fill it in.
 **/
static void identity2HeloHelper(char * fn,
				char * dirName,
				struct TempStorage_ * res) 
{
  if (strstr(fn, (char*)&res->hex.data[0]) != NULL) 
    {
      char * fileName;
      HELO_Message buffer;
      int size;
        
      fileName = xmalloc(strlen(networkIdDirectory) + strlen(fn) + 1, "identity2HeloHelper: filename");
      sprintf(fileName, "%s%s", networkIdDirectory, fn);
      size = readFile(fileName, sizeof(HELO_Message), &buffer);
      if (size == sizeof(HELO_Message)) 
	    {
          HELO_Message * tmp;
          tmp = xmalloc(HELO_Message_size(&buffer), "reading helo from drive (identity2helo)");
          size = readFile(fileName, HELO_Message_size(&buffer), tmp);
          if (size != HELO_Message_size(&buffer)) 
		    {
	          LOG(LOG_WARNING, "WARNING: Removed invalid HELO file %s (filesize %d, expected %d)\n", fileName, size, HELO_Message_size(&buffer));	
	          unlink(fileName);      
	          xfree(tmp, "helo was invalid, freeing buffer");
            } 
		  else 
		    {
	          if (NO == isTransportAvailable(ntohs(buffer.protocol)) ) 
		        {
	              xfree(tmp, "helo was for unsupported transport protocol");
	            } 
		      else if (res->result == SYSERR) 
		        {
	              res->result = OK;
	              res->helo = tmp;
	            } 
		      else if (randomi(transportGetCost(ntohs(tmp->protocol))) > randomi(transportGetCost(ntohs(res->helo->protocol)))) 
		        {
	              xfree(res->helo, "another proto is better"); 
			      res->helo = tmp;
	            } 
		      else
	            xfree(tmp, "existing proto won");
            }
         } 
	  else 
	    {
          LOG(LOG_WARNING, "WARNING File %s not large enough: only %d < %d bytes (removed)\n", fileName, size, sizeof(HELO_Message));
          unlink(fileName);      
        }
      xfree(fileName, "identity2HeloHelper: filename");
    }
}

/**
 * Obtain the public key and address of a known host. If no specific
 * protocol is specified (ANY_PROTOCOL_NUMBER), HELOs for cheaper
 * protocols are returned with preference (randomness!).
 *
 * @param hostId the host id
 * @param protocol the protocol that we need,
 *        ANY_PROTOCOL_NUMBER if we do not care which protocol
 * @param tryTemporaryList is it ok to check the unverified HELOs?
 * @param result where to store the result
 * @returns SYSERR on failure, OK on success
 **/
int identity2Helo(const HostIdentity *  hostId,
		  const unsigned short protocol,
		  int tryTemporaryList,
		  HELO_Message ** result) 
{  
  struct TempStorage_ tempStorage;
  char * fn;
  HELO_Message buffer;
  int size;
  int i;

  *result = NULL;

  /* printf("a=%x b=%x c=%x, d=%x e=%x!\n", hostId->hashPubKey.a, hostId->hashPubKey.b, hostId->hashPubKey.c, hostId->hashPubKey.d, hostId->hashPubKey.e); */

  fn = getHostFileName(hostId, protocol);
  
  /* printf("hostfilename=%s!\n", fn); */

  size = readFile(fn, sizeof(HELO_Message), &buffer);
 
  /* printf("size of file=%d!\n", size); */

  if (size == sizeof(HELO_Message)) 
    {
	    /* printf("pass 0001\n"); */
      *result = xmalloc(HELO_Message_size(&buffer), "reading helo from drive (identity2helo)");
	    /* printf("pass 0002\n"); */
      size = readFile(fn, HELO_Message_size(&buffer), *result);
	    /* printf("pass 0003\n"); */
      if (size != HELO_Message_size(&buffer)) 
	    {
			  /* printf("pass 0004\n"); */
          LOG(LOG_WARNING, "WARNING: Removed invalid HELO file %s\n", fn);
		    /* printf("pass 0005\n"); */
          unlink(fn);      
		    /* printf("pass 0006\n"); */
          xfree(fn, "identity2key: filename (2)");
		    /* printf("pass 0007\n"); */
          xfree(*result, "helo was invalid, freeing buffer");
		    /* printf("pass 0008\n"); */
          *result = NULL;
		    /* printf("pass 0009\n"); */
          return SYSERR;
        }
		  /* printf("pass 0010\n"); */
	  /* printf("isTransportAvailable=%d!\n", isTransportAvailable(ntohs((*result)->protocol)));
	  printf("pass 0012a YES/NO\n"); */
      xfree(fn, "identity2key: filename (2)");
      /* printf("pass 0012\n"); */
      return OK;

      /* if (YES == isTransportAvailable(ntohs((*result)->protocol))) 
	    {
			  printf("pass 0011 YES\n");
          xfree(fn, "identity2key: filename (2)");
		    printf("pass 0012\n");
          return OK;
        } 
	  else 
	    {
			  printf("pass 0013\n");
          xfree(*result, "helo was invalid, freeing buffer");
		    printf("pass 0014\n");
          *result = NULL;
        } */
    } 
  else if (size != -1) 
    {
		  /* printf("pass 0015\n"); */
      LOG(LOG_WARNING, "WARNING: Removed invalid HELO file %s\n", fn);
	    /* printf("pass 0016\n"); */
      unlink(fn);      
	    /* printf("pass 0017\n"); */
    }
	  /* printf("pass 0018\n"); */
  xfree(fn, "identity2key: filename (1)");
  /* printf("pass 0019\n"); */
  if (YES == tryTemporaryList) 
    {
	  /* printf("pass 0020\n"); */
      /* ok, then try temporary hosts */      
      MUTEX_LOCK(&lock_);   
      for (i=0;i<MAX_TEMP_HOSTS;i++) 
	     {
           if ( (tempHosts[i] != NULL) && hostIdentityEquals(hostId, &tempHosts[i]->senderIdentity) && ( (ntohs(tempHosts[i]->protocol) == protocol) ||( (protocol == ANY_PROTOCOL_NUMBER) && (YES == isTransportAvailable(tempHosts[i]->protocol)) ) ) ) 
		     {
	           *result = xmalloc(HELO_Message_size(tempHosts[i]), "found temporary HELO match");
	           memcpy(*result, tempHosts[i], HELO_Message_size(tempHosts[i]));	    
	           MUTEX_UNLOCK(&lock_);   
	           return OK;
             }    
         }
      MUTEX_UNLOCK(&lock_);   
    }
  if (protocol != ANY_PROTOCOL_NUMBER) return SYSERR; /* nothing found */
  /* ok, last chance, scan directory! */
  hash2hex(&hostId->hashPubKey, &tempStorage.hex);

  tempStorage.result = SYSERR;
  tempStorage.helo = NULL;
  LOG(LOG_DEBUG, "DEBUG: scanning directory %s for peer identity, proto %d\n", networkIdDirectory, protocol);
  scanDirectory(networkIdDirectory, (DirectoryEntryCallback)&identity2HeloHelper, &tempStorage);
  *result = tempStorage.helo;

  return tempStorage.result;
}

/**
 * Free memory occupied by hostkey
 **/
void freeHostkey(HOSTKEY hostkey) 
{
  #ifdef USE_OPENSSL
    RSA_free(hostkey);
  #else
    gcry_sexp_release(hostkey);
  #endif
}

/**
 * Internal: publicKey => RSA-Key
 **/
static HOSTKEY public2Hostkey(PublicKey * publicKey) 
{
  #ifdef USE_OPENSSL
    HOSTKEY result;
    int sizen;
    int sizee;

    result = RSA_new();
    if (ntohs(publicKey->len) != sizeof(PublicKey)-sizeof(publicKey->padding)) 
	  {
        LOG(LOG_ERROR, "ERROR: public2Hostkey: received invalid publicKey (size=%d)\n", ntohs(publicKey->len));
        return NULL;
      }
    sizen = ntohs(publicKey->sizen);
    sizee = ntohs(publicKey->len) - sizen - 2*sizeof(unsigned short);
    if ( (sizen != RSA_ENC_LEN) || (sizee + sizen != RSA_KEY_LEN)) 
	  {
        LOG(LOG_ERROR, "ERROR: public2Hostkey: received invalid publicKey (sizee=%d, sizen=%d)\n", sizee, sizen);
        return NULL;
      }
    result->n = BN_bin2bn(&publicKey->key[0], sizen, NULL);
    result->e = BN_bin2bn(&publicKey->key[sizen], sizee, NULL);
    return result;
  #else
    HOSTKEY result;
    GcryMPI n;
    GcryMPI e;
    size_t size;
    size_t erroff;
    int rc;
    /* len = sizen + sizee + 4 == publickey - padding! */
    if ( ( ntohs(publicKey->sizen) != RSA_ENC_LEN ) || ( ntohs(publicKey->len) != sizeof(PublicKey) - sizeof(publicKey->padding)) ) 
	  {
        LOG(LOG_ERROR, "ERROR: public2Hostkey: received invalid publicKey (%d, %d)\n", ntohs(publicKey->len), ntohs(publicKey->sizen)); 
        return NULL;
      }
    size = RSA_ENC_LEN;
    rc = gcry_mpi_scan(&n, GCRYMPI_FMT_USG, &publicKey->key[0], &size);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: gcry_mpi_scan of n failed (%s)\n", gcry_strerror(rc));
        return NULL;
      }
    size = RSA_KEY_LEN - RSA_ENC_LEN;
    rc = gcry_mpi_scan(&e, GCRYMPI_FMT_USG, &publicKey->key[RSA_ENC_LEN], &size);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: gcry_mpi_scan of e failed (%s)\n", gcry_strerror(rc));
        gcry_mpi_release(n);
        return NULL;
      }
    rc = gcry_sexp_build(&result, &erroff, "(public-key(rsa(n %m)(e %m)))", n, e);
    gcry_mpi_release(n);
    gcry_mpi_release(e);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: gcry_sexp_build of public key failed (%s at position %d)\n", gcry_strerror(rc), erroff);
        return NULL;
      }  
    return result;
  #endif
}

/**
 * Verify signature.
 *
 * @param block the signed data
 * @param len the length of the block 
 * @param sig signature
 * @param publicKey public key of the signer
 * @returns OK if ok, SYSERR if invalid
 **/
int verifySig(void * block,
	      unsigned short len,
	      Signature * sig,	      
	      PublicKey * publicKey) 
{
  #ifdef USE_OPENSSL
    HOSTKEY hostkey;
    int rs;
    HashCode160 hc;
 
    hostkey = public2Hostkey(publicKey);
    if ((hostkey == NULL) || (sig == NULL) || (block == NULL)) return SYSERR; /* hey, no data !? */

    rs = RSA_size(hostkey);
    if (rs != RSA_ENC_LEN) 
      {
        errexit("FATAL: verifySig: rs != RSA_ENC_LEN (%d)!", rs);
      }
    hash(block, len, &hc);
    if (1 != RSA_verify(NID_ripemd160, (char*)&hc, sizeof(HashCode160), &sig->sig[0], sizeof(Signature), hostkey)) 
      {
        LOG(LOG_ERROR, "ERROR: verifySig: signature mismatch\n");
        freeHostkey(hostkey);
        return SYSERR;
      }
    freeHostkey(hostkey);
    return OK;
  #else
    /* libgcrypt */
    GcrySexp data;
    GcrySexp sigdata;
    size_t size;
    GcryMPI val;
    HOSTKEY hostkey;
    HashCode160 hc;
    char * buff;
    int bufSize;
    size_t erroff;
    int rc;
 
    size = sizeof(Signature);
    rc = gcry_mpi_scan(&val, GCRYMPI_FMT_USG, (char*)sig, &size);
    if (rc) 
      {
        LOG(LOG_ERROR, "ERROR: sign: key_from_sexp failed (%s)\n", gcry_strerror(rc));
        return SYSERR;
      }
    rc = gcry_sexp_build(&sigdata, &erroff, "(sig-val(rsa(s %m)))", val);
    gcry_mpi_release(val);
    if (rc) 
      {
        LOG(LOG_ERROR, "ERROR: sign: gcry_sexp_build failed (%s)\n", gcry_strerror(rc));
        return SYSERR;
      }  
    hash(block, len, &hc);
    bufSize = strlen(FORMATSTRING) + 1;
    buff = xmalloc(bufSize, "sexp building");
    memcpy(buff, FORMATSTRING, bufSize);
    memcpy(&buff[strlen(FORMATSTRING) - strlen("01234567890123456789))")], &hc, sizeof(HashCode160));
    rc = gcry_sexp_new(&data, buff, bufSize, 0);
    xfree(buff, "sexp building");
    hostkey = public2Hostkey(publicKey);
    rc = gcry_pk_verify(sigdata, data, hostkey);  
    freeHostkey(hostkey);
    gcry_sexp_release(data);
    gcry_sexp_release(sigdata);
    if (rc) 
      {
        LOG(LOG_WARNING, "WARNING: signature verification failed (%s)\n", gcry_strerror(rc));
        return SYSERR;
      } 
    else
      return OK;
  #endif
}

/**
 * Convert HELO to string.
 **/
char * heloToString(HELO_Message * helo) 
{
  TransportAPI * tapi;
  unsigned short prot;
  
  if (ntohs(helo->protocol) >= tapis_count) 
    {
      LOG(LOG_INFO, "INFO: heloToString failed, protocol %d not supported\n",
	  ntohs(helo->protocol));
      return NULL;
    }
  prot = ntohs(helo->protocol);
  tapi = tapis[prot];
  if (tapi == NULL) 
    {
      LOG(LOG_INFO, "INFO: transportConnect failed, protocol %d not supported\n",
	  ntohs(helo->protocol));
      return NULL;
    } 
  else 
    return tapi->addressToString(helo);
}

void printHostInfo(const HostIdentity * id,
			  const unsigned short proto,   
			  void * data) 
{
  HELO_Message * helo;
  char * info;
  HexName hex;

  hash2hex(&id->hashPubKey, &hex);
  if (SYSERR == identity2Helo(id, proto, NO, &helo)) 
    {
      LOG(LOG_WARNING, "WARNING: could not get address of %s\n", &hex);
      return;
    }  
  if (SYSERR == verifySig(&helo->senderIdentity,
			  HELO_Message_size(helo) - sizeof(Signature) - sizeof(PublicKey) - sizeof(p2p_HEADER),
			  &helo->signature,
			  &helo->publicKey)) 
    {
      LOG(LOG_WARNING, "WARNING: HELO message invalid (signature bogous).\n");
    }

  info = heloToString(helo);
 
  xfree(helo, "freeing helo");

  if (info == NULL) 
    {
      LOG(LOG_WARNING, "WARNING: could not get address of %s\n", &hex);
      return;
    }

  printf("%s at address %s\n",(char*)&hex, info);

  xfree(info, "heloToString characters");  
}

int protocol_conversion(const unsigned short proto)
{
  switch (proto) 
	{
		case 6:
			{
			  return 2;
			};break;
		case 17:
			{
			  return 1;
			};break;
	}

  return -1;
}

/**
 * Convert UDP/TCP address to a string.
 **/
char * addressToString(HELO_Message * helo, const unsigned short proto) 
{
  char * ret;
  HostAddress * haddr;
  
  haddr = (HostAddress*) &helo->senderAddress[0];  
  ret = xmalloc(4*4+6+6+2, "buffer for helo-toString");
  sprintf(ret, "'%d.%d.%d.%d',%d,%d", PRIP(ntohl(*(int*)&haddr->senderIP.addr)), ntohs(haddr->senderPort), protocol_conversion(proto));
  return ret;
}

/********************************************************
* prints the statement resultset                        *
*********************************************************/
int my_print_resultset2(SQLHSTMT hstmt, int *protocol)
{
  SQLRETURN   rc;
  SQLUINTEGER nRowCount=0, pcColDef;
  SQLCHAR     szColName[MAX_NAME_LEN];
  SQLCHAR     szData[MAX_COLUMNS][MAX_ROW_DATA_LEN]={""};
  SQLSMALLINT nIndex,ncol,pfSqlType, pcbScale, pfNullable;
  (*protocol) = 0;

    /* get total number of columns from the resultset */
    rc = SQLNumResultCols(hstmt,&ncol);
    mystmt(hstmt,rc);

    /* print the column names  and do the row bind */
    for(nIndex = 1; nIndex <= ncol; nIndex++)
    {
      rc = SQLDescribeCol(hstmt,nIndex,szColName, MAX_NAME_LEN+1, NULL,
                          &pfSqlType,&pcColDef,&pcbScale,&pfNullable);
      mystmt(hstmt,rc);

      /* printf(" %s\t",szColName); */

      rc = SQLBindCol(hstmt,nIndex, SQL_C_CHAR, szData[nIndex-1],
                      MAX_ROW_DATA_LEN+1,NULL);
      mystmt(hstmt,rc);
    }

    /* printf("\n -------------------------------------------\n"); */

    /* now fetch row by row */
    rc = SQLFetch(hstmt);
    while(rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO)
    {
       nRowCount++;
       for(nIndex=0; nIndex< ncol; nIndex++)
	     {
           /* printf(" %s\t",szData[nIndex]); */
		   (*protocol) = atoi(szData[nIndex]);
	       /* printf("protocol=%s==%d!\n", szData[nIndex], (*protocol)); */
		 }
     
       /* printf("\n"); */
       rc = SQLFetch(hstmt);
     }
     SQLFreeStmt(hstmt,SQL_UNBIND);

     /* printf("\n total rows fetched:%d\n", ((int) nRowCount)); */

     /* free the statement row bind resources */
     rc = SQLFreeStmt(hstmt, SQL_UNBIND);
     mystmt(hstmt,rc);

     /* free the statement cursor */
     rc = SQLFreeStmt(hstmt, SQL_CLOSE);
     mystmt(hstmt,rc);   
     
     return(nRowCount);
}

void printHostInfo2(const HostIdentity * id,
			  const unsigned short proto) 
{
  HELO_Message * helo;
  /* char * info; */
  HexName hex;
  char * addressToString_pointer;
  char timebuf[64];
  char input_string[255];
  time_t timetmp;
  struct tm *tmptr;
  SQLRETURN   rc;
  SQLSMALLINT ncol;
  int number_of_rows;
  int protocol = 0, protocol_new = 0;


  time(&timetmp);
  tmptr=localtime(&timetmp);
  strftime(timebuf, 64, "%Y%m%d,\'%H:%M:%S\'", tmptr);

  /* printf("timebuf=%s!\n", timebuf); */

  hash2hex(&id->hashPubKey, &hex);
  if (SYSERR == identity2Helo(id, proto, NO, &helo)) 
    {
      LOG(LOG_WARNING, "WARNING: could not get address of %s\n", &hex);
      return;
    }  
  if (SYSERR == verifySig(&helo->senderIdentity,
			  HELO_Message_size(helo) - sizeof(Signature) - sizeof(PublicKey) - sizeof(p2p_HEADER),
			  &helo->signature,
			  &helo->publicKey)) 
    {
      LOG(LOG_WARNING, "WARNING: HELO message invalid (signature bogous).\n");
    }

  /* info = heloToString(helo);
  printf("info=%s!\n", info); */

  /* if (info == NULL) 
    {
      LOG(LOG_WARNING, "WARNING: could not get address of %s\n", &hex);
      return; 
    } */

  addressToString_pointer = addressToString(helo, proto);
  /* printf("%s at address %s\n",(char*)&hex, info); */

  /* printf("addressToString_pointer=%s!proto=%d!\n", addressToString_pointer, proto); */

  sprintf(input_string, "SELECT protocol FROM fshpeerinfo WHERE hostkey='%s';\n", (char*)&hex);
    /* Now fetch and verify the data */
  rc = SQLExecDirect(hstmt,input_string,SQL_NTS);
  mystmt(hstmt,rc);

  /* get total number of columns from the resultset */
  rc = SQLNumResultCols(hstmt,&ncol);
  mystmt(hstmt,rc);

  /* printf("number of coloms in result=%d!\n", ncol); */

  number_of_rows = my_print_resultset2(hstmt, &protocol);
  /* printf("0123My_print_resultset=%d!\n", number_of_rows); */

  /* printf("INSERT INTO fshpeerinfo VALUES (%s,'%s',%s);\n", timebuf, (char*)&hex, addressToString_pointer); */

  if (number_of_rows == 0)
    {
      sprintf(input_string, "INSERT INTO fshpeerinfo VALUES (%s,'%s',%s);\n", timebuf, (char*)&hex, addressToString_pointer);

      /* printf("input_string=%s!\n", input_string); */
      /* insert in the table 'fshpeerinfo' */
      rc = SQLExecDirect(hstmt,input_string,SQL_NTS);

      mystmt(hstmt,rc);

      /* commit the transaction*/
      rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT);
      mycon(hdbc,rc);
	}
  else
    {
      if (number_of_rows == 1)
	    {
		   protocol_new = (protocol_conversion(proto) | protocol) & 0x07;
           /* printf("proto=%d protocol=%d protocol_new=%d!\n", protocol_conversion(proto), protocol, protocol_new); */
		   if (protocol != protocol_new)
		     {
			   sprintf(input_string, "UPDATE fshpeerinfo set protocol = %d WHERE hostkey='%s';\n", protocol_new, (char*)&hex);
			   /* prepare the insert statement with parameters */
			   /* printf("update_string=%s!\n", input_string); */
			   rc = SQLExecDirect(hstmt,input_string,SQL_NTS);
          
               mystmt(hstmt,rc);
			   /* commit the transaction*/
               rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT);
               mycon(hdbc,rc);
		     }
		}
    }
  xfree(addressToString_pointer, "freeing addressToString_pointer");

  xfree(helo, "freeing helo");

  /* xfree(info, "heloToString characters");   */
}

/**
 * Call a method for each known host.
 *
 * @param callback the method to call for each host
 * @param now the time to use for excluding hosts due to blacklisting, use 0 
 *        to go through all hosts.
 * @param data an argument to pass to the method
 * @return the number of hosts matching
 **/
int forEachHost(HostIterator callback,
		cron_t now,
		void * data) 
{
  int i;
  int count = 0;

  MUTEX_LOCK(&lock_);   

  for (i=0;i<count_;i++) 
     {
       if (hostIdentityEquals(&hosts_[i].identity, &myIdentity)) continue;
       if ( (now == 0) || (now >= hosts_[i].until) ) 
	     {
           count++;
           if (callback != NULL) 
		     {
	           HostIdentity hi;
	           unsigned short proto;

	           memcpy(&hi, &hosts_[i].identity, sizeof(HostIdentity));
	           proto = hosts_[i].protocol;
	           MUTEX_UNLOCK(&lock_);   

               /* printf("forEachHost a=%x b=%x c=%x, d=%x e=%x!\n", hi.hashPubKey.a, hi.hashPubKey.b, hi.hashPubKey.c, hi.hashPubKey.d, hi.hashPubKey.e); */

			   /* printf("proto = %d!\n", proto); */

	           callback(&hi, proto, data);      
	           MUTEX_LOCK(&lock_);   
             }
         }
     }
  MUTEX_UNLOCK(&lock_);   
  return count;
}

/**
 * Call a method for each known host.
 *
 * @param callback the method to call for each host
 * @param now the time to use for excluding hosts due to blacklisting, use 0 
 *        to go through all hosts.
 * @param data an argument to pass to the method
 * @return the number of hosts matching
 **/
int forEachHost2() 
{
  int i;
  int count = 0;

  MUTEX_LOCK(&lock_);   

  for (i=0;i<count_;i++) 
     {
       if (hostIdentityEquals(&hosts_[i].identity, &myIdentity)) continue;
	     {
           count++;
      
	       HostIdentity hi;
	       unsigned short proto;

	       memcpy(&hi, &hosts_[i].identity, sizeof(HostIdentity));
	       proto = hosts_[i].protocol;
	       MUTEX_UNLOCK(&lock_);   

           /* printf("forEachHost a=%x b=%x c=%x, d=%x e=%x!\n", hi.hashPubKey.a, hi.hashPubKey.b, hi.hashPubKey.c, hi.hashPubKey.d, hi.hashPubKey.e); */

	       /* printf("proto = %d!\n", proto); */

	       printHostInfo2(&hi, proto);      
	       MUTEX_LOCK(&lock_);   
             
         }
     }
  MUTEX_UNLOCK(&lock_);   
  return count;
}
		
/**
 * Initialize Random number generator.
 **/

void initRAND() 
{
  #ifndef USE_OPENSSL
    gcry_control(GCRYCTL_DISABLE_SECMEM, 0);
    if (! gcry_check_version(GCRYPT_VERSION)) errexit("FATAL: libgcrypt has not the exptected version %s\n", GCRYPT_VERSION);
  #endif
  srand((unsigned int)time(NULL));
}

/**
 * This HostKey implementation uses RSA.
 **/
HOSTKEY makeHostkey() 
{
  #ifdef USE_OPENSSL
    return RSA_generate_key(HOSTKEY_LEN, 65535, NULL, 0);
  #else
    HOSTKEY s_key;
    GCRY_SEXP s_keyparam;
    int rc;

    rc = gcry_sexp_build(&s_keyparam, NULL, "(genkey(rsa(nbits %d)(rsa-use-e 3:257)))", HOSTKEY_LEN);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: failed to convert keyparam: %s\n", gcry_strerror (rc));
        return NULL;
      }
    rc = gcry_pk_genkey(&s_key, s_keyparam);
    gcry_sexp_release(s_keyparam);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: key generation failed: %s\n", gcry_strerror (rc));
        return NULL;
      }

    #if EXTRA_CHECKS
      if (gcry_pk_testkey(s_key)) 
	    {
          LOG(LOG_ERROR, "ERROR: freshly created hostkey is not sane!\n");
          return NULL;
        }
    #endif
    return s_key;
  #endif
}

/**
 * Encode the private key in a format suitable for
 * storing it into a file.
 * @returns encoding of the private key.
 *    The first 4 bytes give the size of the array, as usual.
 **/
HostKeyEncoded * encodeHostkey(HOSTKEY hostkey) 
{
  #ifdef USE_OPENSSL
    /*
               BIGNUM *n;               public modulus
               BIGNUM *e;               public exponent
               BIGNUM *d;               private exponent
               BIGNUM *p;               secret prime factor
               BIGNUM *q;               secret prime factor
               BIGNUM *dmp1;            d mod (p-1)
               BIGNUM *dmq1;            d mod (q-1)
               BIGNUM *iqmp;            q^-1 mod p
    */
    unsigned short sizen;
    unsigned short sizee;
    unsigned short sized;
    unsigned short sizep;
    unsigned short sizeq;
    unsigned short sizedmp1;
    unsigned short sizedmq1;
    unsigned short sizeiqmp;
    unsigned short size;
    HostKeyEncoded * retval;

    sizen = BN_num_bytes(hostkey->n);
    sizee = BN_num_bytes(hostkey->e);
    sized = BN_num_bytes(hostkey->d);
    if (hostkey->p != NULL) sizep = BN_num_bytes(hostkey->p);
    else
      sizep = 0;
    if (hostkey->q != NULL) sizeq = BN_num_bytes(hostkey->q);
    else 
	  sizeq = 0;
    if (hostkey->dmp1 != NULL) sizedmp1 = BN_num_bytes(hostkey->dmp1);
    else
      sizedmp1 = 0;
    if (hostkey->dmq1 != NULL) sizedmq1 = BN_num_bytes(hostkey->dmq1);
    else
      sizedmq1 = 0;
    if (hostkey->iqmp != NULL) sizeiqmp = BN_num_bytes(hostkey->iqmp);
    else
      sizeiqmp = 0;
    size = sizen+sizee+sized+sizep+sizeq+sizedmp1+sizedmq1+sizeiqmp+sizeof(HostKeyEncoded);
    retval = (HostKeyEncoded *) xmalloc(size, "encodeHostkey: hostkey");
    retval->len = htons(size);
    retval->sizen = htons(sizen);
    retval->sizee = htons(sizee);
    retval->sized = htons(sized);
    retval->sizep = htons(sizep);
    retval->sizeq = htons(sizeq);
    retval->sizedmp1 = htons(sizedmp1);
    retval->sizedmq1 = htons(sizedmq1);
    BN_bn2bin(hostkey->n, &retval->key[0]);
    BN_bn2bin(hostkey->e, &retval->key[0+sizen]);
    BN_bn2bin(hostkey->d, &retval->key[0+sizen+sizee]);
    if (hostkey->p != NULL)BN_bn2bin(hostkey->p, &retval->key[0+sizen+sizee+sized]);
    if (hostkey->q != NULL)BN_bn2bin(hostkey->q, &retval->key[0+sizen+sizee+sized+sizep]);
    if (hostkey->dmp1 != NULL)BN_bn2bin(hostkey->dmp1, &retval->key[0+sizen+sizee+sized+sizep+sizeq]);
    if (hostkey->dmq1 != NULL)BN_bn2bin(hostkey->dmq1, &retval->key[0+sizen+sizee+sized+sizep+sizeq+sizedmp1]);
    if (hostkey->iqmp != NULL)BN_bn2bin(hostkey->iqmp, &retval->key[0+sizen+sizee+sized+sizep+sizeq+sizedmp1+sizedmq1]);
    return retval;
  #else
    /* libgcrypt */

    HostKeyEncoded * retval;
    GcryMPI pkv[6];
    void * pbu[6];
    size_t sizes[6];
    int rc;
    int i;
    int size;

    #if EXTRA_CHECKS
      if (gcry_pk_testkey(hostkey)) 
	    {
          LOG(LOG_ERROR, "ERROR: encodeHostkey: hostkey is not sane!\n");
          return NULL;
        }
    #endif

    memset(pkv, 0, sizeof(GcryMPI) * 6);
    rc = key_from_sexp(pkv, hostkey, "private-key", "nedpqu");
    if (rc) rc = key_from_sexp(pkv, hostkey, "rsa", "nedpqu");
    if (rc) rc = key_from_sexp(pkv, hostkey, "private-key", "nedpq");
    if (rc) rc = key_from_sexp(pkv, hostkey, "rsa", "nedpq");
    if (rc) rc = key_from_sexp(pkv, hostkey, "private-key", "ned");
    if (rc) rc = key_from_sexp(pkv, hostkey, "rsa", "ned");
    if (rc) 
      {
        LOG(LOG_ERROR, "ERROR: encodeHostkey: key_from_sexp failed: %s\n", gcry_strerror (rc));
        return NULL;
      }
    size = sizeof(HostKeyEncoded);
    for (i=0;i<6;i++) 
       {
         if (pkv[i] != NULL)  
	       {
             rc = gcry_mpi_aprint(GCRYMPI_FMT_USG, &pbu[i], &sizes[i], pkv[i]);
             size += sizes[i];
             if (rc)  
		       {
	             LOG(LOG_ERROR, "ERROR: gcry_mpi_aprint failed: %s\n", gcry_strerror (rc));
	             while (i>0) 
			       {
	  			     freeNonNull(pbu[--i], "key conversion buffer");
	               }
	             for (i=0;i<6;i++)
	                if (pkv[i] != NULL) gcry_mpi_release(pkv[i]);
	             return NULL;
               }
           } 
	     else 
	       {
             pbu[i] = NULL;
             sizes[i] = 0;
           }
       }
    if (size >= 65536) errexit("FATAL: size of serialized private key >= 64k\n");  
    retval = xmalloc(size, "allocting host key encoded");
    retval->len = htons(size);
    i = 0;
    retval->sizen = htons(sizes[0]);
    memcpy(&retval->key[i], pbu[0], sizes[0]);
    i += sizes[0];
    retval->sizee = htons(sizes[1]);
    memcpy(&retval->key[i], pbu[1], sizes[1]);
    i += sizes[1];
    retval->sized = htons(sizes[2]);
    memcpy(&retval->key[i], pbu[2], sizes[2]);
    i += sizes[2];
    /* swap p and q! */
    retval->sizep = htons(sizes[4]);
    memcpy(&retval->key[i], pbu[4], sizes[4]);
    i += sizes[4];
    retval->sizeq = htons(sizes[3]);
    memcpy(&retval->key[i], pbu[3], sizes[3]);
    i += sizes[3];
    retval->sizedmp1 = htons(0);
    retval->sizedmq1 = htons(0);
    memcpy(&retval->key[i], pbu[5], sizes[5]);
    for (i=0;i<6;i++) 
       {
         if (pkv[i] != NULL) gcry_mpi_release(pkv[i]);
         freeNonNull(pbu[i], "freeing temporaries in hostkey serialization");
       }
    return retval;
  #endif
}

/**
 * Decode the private key from the file-format back
 * to the "normal", internal format.
 **/
HOSTKEY decodeHostkey(HostKeyEncoded * encoding) 
{
  #ifdef USE_OPENSSL
    unsigned short sizen;
    unsigned short sizee;
    unsigned short sized;
    unsigned short sizep;
    unsigned short sizeq;
    unsigned short sizedmp1;
    unsigned short sizedmq1;
    unsigned short size;
    unsigned short sum;

    HOSTKEY result = (HOSTKEY) RSA_new();
    size    = ntohs(encoding->len) - sizeof(HostKeyEncoded);
    sizen   = ntohs(encoding->sizen);
    sizee   = ntohs(encoding->sizee);
    sized   = ntohs(encoding->sized);
    sizep   = ntohs(encoding->sizep);
    sizeq   = ntohs(encoding->sizeq);
    sizedmp1= ntohs(encoding->sizedmp1);
    sizedmq1= ntohs(encoding->sizedmq1);
    sum = 0;
    result->n= BN_bin2bn(&encoding->key[sum], sizen, NULL); sum += sizen;
    result->e= BN_bin2bn(&encoding->key[sum], sizee, NULL); sum += sizee;
    result->d= BN_bin2bn(&encoding->key[sum], sized, NULL); sum += sized;
    if (sizep != 0) 
	  {
        result->p = BN_bin2bn(&encoding->key[sum], sizep, NULL); sum += sizep;
      } 
	else
      result->p = NULL;
    if (sizeq != 0) 
	  {
        result->q = BN_bin2bn(&encoding->key[sum], sizeq, NULL); sum += sizeq;
      } 
	else
      result->q = NULL;
    if (sizedmp1 != 0) 
	  {
        result->dmp1= BN_bin2bn(&encoding->key[sum], sizedmp1, NULL); sum += sizedmp1;
      } 
	else
      result->dmp1 = NULL;
    if (sizedmq1 != 0) 
	  {
        result->dmq1 = BN_bin2bn(&encoding->key[sum], sizedmq1, NULL); sum += sizedmq1;
      } 
	else
      result->dmq1 = NULL;
    if (size - sum > 0) result->iqmp= BN_bin2bn(&encoding->key[sum], size-sum, NULL);
    else
      result->iqmp = NULL;
    return result;
  #else
    HOSTKEY res;
    GcryMPI n,e,d,p,q,u;
    int rc;
    size_t size;
    int pos;

    pos = 0;
    size = ntohs(encoding->sizen);
    rc = gcry_mpi_scan(&n, GCRYMPI_FMT_USG, &encoding->key[pos], &size);
    pos += ntohs(encoding->sizen);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s)\n", gcry_strerror(rc));
        return NULL;
      }
    size = ntohs(encoding->sizee);
    rc = gcry_mpi_scan(&e, GCRYMPI_FMT_USG, &encoding->key[pos], &size);
    pos += ntohs(encoding->sizee);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s)\n", gcry_strerror(rc));
        gcry_mpi_release(n);
        return NULL;
      }
    size = ntohs(encoding->sized);
    rc = gcry_mpi_scan(&d, GCRYMPI_FMT_USG, &encoding->key[pos], &size);
    pos += ntohs(encoding->sized);
    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s)\n", gcry_strerror(rc));
        gcry_mpi_release(n);
        gcry_mpi_release(e);
        return NULL;
      }
    /* swap p and q! */
    size = ntohs(encoding->sizep);
    if (size > 0) 
	  {
        rc = gcry_mpi_scan(&q, GCRYMPI_FMT_USG, &encoding->key[pos], &size);
        pos += ntohs(encoding->sizep);
        if (rc) 
		  {
            LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s)\n", gcry_strerror(rc));
            gcry_mpi_release(n);
            gcry_mpi_release(e);
            gcry_mpi_release(d);
            return NULL;
          }
      } 
	else
      q = NULL;
    size = ntohs(encoding->sizeq);
    if (size > 0) 
	  {
        rc = gcry_mpi_scan(&p, GCRYMPI_FMT_USG, &encoding->key[pos], &size);
        pos += ntohs(encoding->sizeq);
        if (rc) 
		  {
            LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s)\n", gcry_strerror(rc));
            gcry_mpi_release(n);
            gcry_mpi_release(e);
            gcry_mpi_release(d);
            if (q != NULL) gcry_mpi_release(q);
            return NULL;
          }
      } 
	else
      p = NULL;
    pos += ntohs(encoding->sizedmp1);
    pos += ntohs(encoding->sizedmq1);
  
    size = ntohs(encoding->len) - sizeof(HostKeyEncoded) - pos;
    if (size > 0) 
	  {
        rc = gcry_mpi_scan(&u,
		       GCRYMPI_FMT_USG,
		       &encoding->key[pos],
		       &size);
        if (rc) 
		  {
            LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s)\n", gcry_strerror(rc));
            gcry_mpi_release(n);
            gcry_mpi_release(e);
            gcry_mpi_release(d);
            if (p != NULL) gcry_mpi_release(p);
            if (q != NULL)gcry_mpi_release(q);
            return NULL;
          }
      } 
	else
      u = NULL;

    if ( (p != NULL) && (q != NULL) && (u != NULL) ) 
	  {
        rc = gcry_sexp_build(&res, &size, /* erroff */ "(private-key(rsa(n %m)(e %m)(d %m)(p %m)(q %m)(u %m)))", n, e, d, p, q, u);
      } 
	else 
	  {
        if ( (p != NULL) && (q != NULL) ) 
		  {
            rc = gcry_sexp_build(&res, &size, /* erroff */ "(private-key(rsa(n %m)(e %m)(d %m)(p %m)(q %m)))", n, e, d, p, q);
          } 
		else 
		  {
            rc = gcry_sexp_build(&res, &size, /* erroff */ "(private-key(rsa(n %m)(e %m)(d %m)))", n, e, d);
          }
      }
    gcry_mpi_release(n);
    gcry_mpi_release(e);
    gcry_mpi_release(d);
    if (p != NULL) gcry_mpi_release(p);
    if (q != NULL) gcry_mpi_release(q);
    if (u != NULL) gcry_mpi_release(u);

    if (rc) 
	  {
        LOG(LOG_ERROR, "ERROR: could not decode hostkey (%s at %d)\n", gcry_strerror(rc), size);
      }
    #if EXTRA_CHECKS
       if (gcry_pk_testkey(res)) 
	     {
           LOG(LOG_ERROR, "ERROR: decodeHostkey: hostkey is not sane!\n");
           return NULL;
         }
    #endif
    return res;
  #endif
}

/**
 * Convert string to value ('755' for chmod-call)
 **/
static int atoo(char *s) 
{
  int n = 0;

  while ( ('0' <= *s) && (*s < '8') ) 
    {
      n <<= 3;
      n += *s++ - '0';
    }
  return n;
}

/**
 * Write a buffer to a file.
 * @param fileName the name of the file, NOT freed!
 * @param buffer the data to write
 * @param n number of bytes to write
 * @param mode permissions to set on the file
 **/ 
void writeFile(char * fileName, 
	       void * buffer,
	       int n,
	       char *mode) 
{
  int handle;
  /* open file, open with 600, create if not 
     present, otherwise overwrite */  
  if ((fileName == NULL) || (buffer == NULL))return;
  handle = open(fileName,O_CREAT|O_WRONLY,S_IRUSR|S_IWUSR);
  /* write the buffer take length from the beginning */
  if (n != write(handle, buffer, n)) LOG(LOG_WARNING, "WARNING: Writing %d bytes to file %s failed!\n", n, fileName);
  chmod(fileName, atoo(mode));
  close(handle);
}

/**
 * Extract the public key of the host.
 * @param hostkey the hostkey to extract into the result.
 * @param result where to write the result.
 **/
void getPublicKey(HOSTKEY hostkey,
		  PublicKey * result) 
{
  #ifdef USE_OPENSSL
    unsigned short sizen;
    unsigned short sizee;
    unsigned short size;

    sizen = BN_num_bytes(hostkey->n);
    sizee = BN_num_bytes(hostkey->e);
    size = sizen + sizee+2*sizeof(unsigned short);
    if (size != sizeof(PublicKey)-sizeof(result->padding)) errexit("FATAL: sizeof public key does not match size (%d!=%d)\n", size, sizeof(PublicKey)-sizeof(result->padding));
    if (RSA_KEY_LEN != sizen+sizee) errexit("FATAL: PublicKey datastructure wrong (%d+%d!=%d)!\n", sizen, sizee, RSA_KEY_LEN);
    result->len = htons(size);
    result->sizen = htons(sizen);
    result->padding = 0;
    BN_bn2bin(hostkey->n, &result->key[0]);
    BN_bn2bin(hostkey->e,&result->key[sizen]);
  #else
    GcryMPI skey[2];
    int size;
    int rc;
  
    rc = key_from_sexp(skey, hostkey, "public-key", "ne");
    if (rc) rc = key_from_sexp(skey, hostkey, "private-key", "ne");    
    if (rc) rc = key_from_sexp(skey, hostkey, "rsa", "ne");    
    if (rc) errexit("FATAL: getPublicKey: key_from_sexp failed: %s\n", gcry_strerror(rc));
  
    result->len = htons(sizeof(PublicKey) - sizeof(result->padding));
    result->sizen = htons(RSA_ENC_LEN);
    result->padding = 0;
    size = RSA_ENC_LEN;
    rc = gcry_mpi_print(GCRYMPI_FMT_USG, &result->key[0], &size, skey[0]);
    if (rc) errexit("FATAL: gcry_mpi_print of n failed: %s\n", gcry_strerror(rc));
    if (size != RSA_ENC_LEN) errexit("FATAL: gcry_mpi_print assert (n) failed! (%d != %d)\n", size, RSA_ENC_LEN);

    size = RSA_KEY_LEN - RSA_ENC_LEN; 
    rc = gcry_mpi_print(GCRYMPI_FMT_USG, &result->key[RSA_ENC_LEN], &size, skey[1]);
    if (rc) errexit("FATAL: gcry_mpi_print of e failed: %s\n", gcry_strerror(rc));
    if (size != RSA_KEY_LEN - RSA_ENC_LEN) errexit("FATAL: gcry_mpi_print assert (e) failed! (%d != %d)\n", size, RSA_KEY_LEN - RSA_ENC_LEN);
  #endif
}

/**
 * Obtain identity from publicHostkey.
 * @param pubKey the public key of the host
 * @param result address where to write the identity of the node
 **/
void getHostIdentity(PublicKey * pubKey,
		     HostIdentity * result) {
  hash(pubKey,
       sizeof(PublicKey),
       &result->hashPubKey);
}

/**
 * Iterate over all available transport mechanisms.
 * @param callback the method to call on each transport API implementation
 * @param data second argument to callback
 **/
void forEachTransport(TransportCallback callback,
		      void * data) 
{
  int i;
  
  for (i=0;i<tapis_count;i++)
    if (tapis[i] != NULL)
      callback(tapis[i], data);
}

/**
 * Get the public key of the host
 * @return reference to the public key. Do not free it!
 **/
PublicKey * getPublicHostkey() 
{
  return publicKey;
}

/**
 * Sign a given block.
 *
 * @param hostkey the hostkey with which to sign this block
 * @param size how many bytes to sign
 * @param block the data to sign
 * @param result where to write the signature
 * @return SYSERR on error, OK on success
 **/
int sign(HOSTKEY hostkey, 
	 unsigned short size,
	 void * block,
	 Signature * sig) 
{
  #ifdef USE_OPENSSL
    #if EXTRA_CHECKS
      PublicKey pkey;
    #endif
    int rs = RSA_size(hostkey);
    int sigSize;
    HashCode160 hc;

    if (block == NULL) return SYSERR;
    if (rs != sizeof(Signature)) 
	  {
        LOG(LOG_ERROR, "ERROR: sign: signature length (RSA_size) has unexpected value (%d)!", rs);
        return SYSERR;
      }
    hash(block, size, &hc);
    if (1 != RSA_sign(NID_ripemd160, (char*)&hc, sizeof(HashCode160), &sig->sig[0], &sigSize, hostkey)) 
	  {
        LOG(LOG_ERROR, "ERROR: RSA_sign failed!", sigSize);
        return SYSERR;
      }
    if (sigSize != sizeof(Signature)) 
	  {
        LOG(LOG_ERROR, "ERROR: sign: sigSize wrong (%u)!", sigSize);
        return SYSERR;
      }
    #if EXTRA_CHECKS
      getPublicKey(hostkey, &pkey);
      if (SYSERR == verifySig(block, size, sig, &pkey)) 
	    {
          errexit("FATAL: sign: generated signature does not pass verification!\n");
        }
    #endif
      return OK;
  #else
     /* libgcrypt */
     GcrySexp result;
     GcrySexp data;
     size_t ssize;
     GcryMPI rval;
     HashCode160 hc;
     char * buff;
     int bufSize;
     int rc;

     hash(block, size, &hc);

     #define FORMATSTRING "(4:data(5:flags5:pkcs1)(4:hash6:rmd16020:01234567890123456789))"

     bufSize = strlen(FORMATSTRING) + 1;
     buff = xmalloc(bufSize, "sexp building");
     memcpy(buff, FORMATSTRING, bufSize);
     memcpy(&buff[bufSize - strlen("012345678901234567890))")], &hc, sizeof(HashCode160));
     rc = gcry_sexp_new(&data, buff, bufSize, 0);
     xfree(buff, "sexp building buffer");
     if (rc) 
	   {
         LOG(LOG_ERROR, "ERROR: sign: gcry_sexp_new failed (%s)\n", gcry_strerror(rc));
         return SYSERR;
       }
     rc = gcry_pk_sign(&result, data, hostkey);
     gcry_sexp_release(data);
     if (rc) 
	   {
         LOG(LOG_ERROR, "ERROR: sign: gcry_pk_sign failed (%s)\n", gcry_strerror(rc));
         return SYSERR;
       }
     rc = key_from_sexp(&rval, result, "rsa", "s");
     gcry_sexp_release(result);
     if (rc) 
	   {
         LOG(LOG_ERROR, "ERROR: sign: key_from_sexp failed (%s)\n", gcry_strerror(rc));
         return SYSERR;
       }
     ssize = sizeof(Signature);
     rc = gcry_mpi_print(GCRYMPI_FMT_USG, (char*)sig, &ssize, rval);
     gcry_mpi_release(rval);
     if (rc) 
	   {
         LOG(LOG_ERROR, "ERROR: sign: gcry_mpi_print failed (%s)\n", gcry_strerror(rc));
         return SYSERR;
       }
     return OK;
  #endif
}

/** 
 * Sign arbitrary data. ALWAYS use only on data we generated
 * entirely! 
 * @return SYSERR on error, OK on success
 **/
int signData(void * data,
	     unsigned short size,
	     Signature * result) 
{
  return sign(hostkey, 
	      size, 
	      data, 
	      result);
}

/**
 * Add a host to the list.
 * @param identity the identity of the host
 * @param protocol the protocol for the host
 **/
static void addHostToKnown(HostIdentity * identity,
			   unsigned short protocol) 
{
  int i;

  MUTEX_LOCK(&lock_);   
  for (i=0;i<count_;i++)
    if ( (hostIdentityEquals(identity, &hosts_[i].identity)) && (protocol == hosts_[i].protocol) ) 
	  {
        MUTEX_UNLOCK(&lock_);   
        return; /* already there */
      }
  if (count_ == max_) 
    { 
      /* grow */ HostEntry * tmp = xmalloc(sizeof(HostEntry) * max_ * 2, "addHostToKnown: grow hostlist");
      memcpy(tmp, hosts_, sizeof(HostEntry) * max_);
      xfree(hosts_, "addHostToKnown: free old hostlist");
      hosts_ = tmp;     
	  max_ = max_ * 2;
     }
  memcpy(&hosts_[count_].identity, identity, sizeof(HostIdentity));
  hosts_[count_].until = 0;
  hosts_[count_].delta = 0;
  hosts_[count_].protocol = protocol;
  count_++;
  MUTEX_UNLOCK(&lock_);   
}

/**
 * Initialize KeyService. Configuration must be initialized at this
 * point. You must call this method first!
 **/
void initKeyService() 
{
  /* char * hostkeyfile; */
  HostKeyEncoded * encHostkey;
  unsigned short len;
  int res;
  HexName myself;

  initRAND();

  /* hostkeyfile = getFileName("GNUNETD", "HOSTKEY", "Configuration file must specify filename for hostkey in section %s under %s.\n");   */

  res = readFile(HOSTKEYFILE, sizeof(unsigned short), &len);

  if (res == sizeof(unsigned short)) 
    {
      encHostkey = (HostKeyEncoded*) xmalloc(ntohs(len), "initKeyservice: hostkeyencoded");    
      if (ntohs(len) != readFile(HOSTKEYFILE, ntohs(len), encHostkey)) 
	    {
          xfree(encHostkey, "initKeyService: hostkeyencoded");
          LOG(LOG_WARNING, "WARNING: hostkey file failed format check, creating new hostkey\n");
          encHostkey = NULL;
        }
    } 
  else
    encHostkey = NULL;

  if (encHostkey == NULL) 
    { /* make new hostkey */
      LOG(LOG_MESSAGE, "Creating new hostkey (this may take a while)...\n");
      hostkey = makeHostkey();
      if (hostkey == NULL)errexit("FATAL: could not create hostkey\n");
      encHostkey = encodeHostkey(hostkey);
      if (encHostkey == NULL) errexit("FATAL: encode hostkey failed\n");
      writeFile(HOSTKEYFILE, encHostkey, ntohs(encHostkey->len), "600");
      LOG(LOG_MESSAGE, "Done creating hostkey.\n");
    } 
  else 
    {
      hostkey = decodeHostkey(encHostkey);    
    }

  /* xfree(hostkeyfile,"filename of the hostkey"); */
  xfree(encHostkey, "initKeyservice: encoded hostkey (2)"); /* ok on encodeHostKey path! */
  publicKey = xmalloc(sizeof(PublicKey), "initKeyservice: encoded public key");
  getPublicKey(hostkey, publicKey);
  getHostIdentity(publicKey, &myIdentity);  
  IFLOG(LOG_DEBUG, hash2hex(&myIdentity.hashPubKey, &myself));
  LOG(LOG_DEBUG, "DEBUG: I am %s\n", &myself);
  /* printf("initKeyService() begin count=%d!\n", count_);
  forEachTransport(&initHelper, NULL);
  printf("initKeyService() end count=%d!\n", count_); */
}

void doneKeyService() 
{
  xfree(publicKey, "doneKeyService: encoded public key");
  freeHostkey(hostkey);
}

/**
 * Get the directory in which we store the hostkeys/HELOs.
 **/
static char * getHostsDirectory() 
{
  strcpy(hostsdirectory, HOSTDIR);

  return hostsdirectory;

  /* return getFileName("GNUNETD",
		     "HOSTS",
		     "Configuration file must specify directory for "\
		     "network identities in section %s under %s.\n"); */
}

/**
 * Complete filename (a la shell) from abbrevition.
 * @param fil the name of the file, may contain ~/ or 
 *        be relative to the current directory
 * @returns the full file name, 
 *          NULL is returned on error
 **/
char * expandFileName(const char * fil) {
  char * fm;
  char * fn;
  const char *fil_ptr;

  if (fil == NULL)
    return NULL;
  
  if (fil[0] == '/') {
    /* absolute path, just copy */
    return(strdup(fil));
  }
  else if (fil[0] == '~') {
    fm = getenv("HOME");
    if (fm == NULL) {
      /* keep it symbolic to show error to user! */
      fm = "$HOME";
    }

    /* do not copy '~' */
    fil_ptr = fil + 1;

	/* skip over dir seperator to be consistent */    
    if (fil_ptr[0] == '/') {
    	fil_ptr++;
    }
  } else {
    fil_ptr = fil;
    fm = getenv("PWD");
    if (fm == NULL)
      fm = "$PWD";
  }

  fn = xmalloc(strlen(fm) + 1 + strlen(fil_ptr) + 1,
    /*dbg */"expandFileName: ");

  sprintf(fn, "%s/%s", fm, fil_ptr);

  return fn;
}

/**
 * Implementation of "mkdir -p"
 * @param dir the directory to create
 **/
void mkdirp(char * dir) 
{
  char * rdir;
  int len;
  int pos;

  rdir = expandFileName(dir); /* expand directory */
  len = strlen(rdir);
  pos = 0;
  while (pos < len) 
    {
      if (rdir[pos] == '/') 
	    {
          rdir[pos] = 0;
          mkdir(rdir,S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH); /* 755 */
          rdir[pos] = '/';       
        }      
      pos++;
    }   
  xfree(rdir, "mkdirp: expandedDirName");
}

/**
 * Convert hex (filename) to the hostIdentity
 * @param hex the filename
 * @param hash is set to the correspoinding host identity
 **/
void hex2hash(HexName * hex,
	      HashCode160 * hash) 
{
  int i;
  int j;
  unsigned char c;
  unsigned char clow;
  unsigned char chigh;

  if ((hex == NULL) || (hash == NULL)) errexit("hex2hash called with hax or hash NULL!");  
  if (strlen((char*)hex) != sizeof(HashCode160)*2) errexit("assertion failed: strlen(hex) is not %d\n", sizeof(HashCode160)*2);
  
  j=0;
  i=0;
  while (i<sizeof(HashCode160)*2) 
    {
      clow = hex->data[i++];
      if (clow >= 'A') clow = clow - 'A' + 10;
      else
        clow = clow - '0';
      chigh = hex->data[i++];    
      if (chigh >= 'A') chigh = chigh - 'A' + 10;
      else
        chigh = chigh - '0';
      c = clow + (chigh << 4);
      ((unsigned char *)hash)[j++] = c;
    }  
}

static void cronHelper(char * filename, 
		       char * dirname,
		       void * unused) 
{
  HostIdentity identity;
  HexName id;
  unsigned int protoNumber;

  if (2 == sscanf(filename, "%40c.%u", (char*)&id, &protoNumber)) 
    {
      ((char*)&id)[40] = '\0';
      hex2hash(&id, &identity.hashPubKey);
	  /* printf("addHostToKnown -> cronHelper count=%d begin!\n", count_); */
      addHostToKnown(&identity, (unsigned short) protoNumber);
	  /* printf("addHostToKnown -> cronHelper count=%d end!\n", count_); */
    } 
  else 
    {
      char * fullname;

      LOG(LOG_WARNING, "WARNING: File in %s does not match naming convention: %s (removed)\n",
	  networkIdDirectory, filename);

      fullname = xmalloc(strlen(filename) + strlen(networkIdDirectory) + 1, "cronHelper removes bad file");
      fullname[0] = '\0';
      strcat(fullname, networkIdDirectory);
      strcat(fullname, filename);
      unlink(fullname);
      xfree(fullname, "cronHelper removes bad file");
    }
}


/**
 * Call this method periodically to scan data/hosts for new hosts.
 **/
void cronScanDirectoryDataHosts(void * unused) 
{
  int count;

  LOG(LOG_CRON, "CRON: enter cronScanDirectoryDataHosts\n");
  count = scanDirectory(networkIdDirectory, &cronHelper, NULL);
  if (count <= 0) 
    {
      LOG(LOG_WARNING, "WARNING: scanDirectory %s returned no known hosts!\n", networkIdDirectory);
    }
  LOG(LOG_CRON, "CRON: exit cronScanDirectoryDataHosts\n");
}

/**
 * Get the current time (works just as "time", just that we use the
 * unit of time that the cron-jobs use).
 *
 * @param setme will set the current time if non-null
 * @return the current time 
 **/
cron_t cronTime(cron_t * setme) 
{
  cron_t res;
  struct timeval tv;
  struct timezone tz; /* man page says it's obsolete, but
			 I'd rather not pass a NULL pointer */
  
  gettimeofday(&tv, &tz);
  res = (((cron_t)tv.tv_sec) * CRON_UNIT_TO_SECONDS) + (tv.tv_usec / NANOSEC_TO_CRON_UNIT);    
  if (setme != NULL) *setme = res;
  return res;
}

/**
 * Add a cron-job to the delta list.
 * @param method which method should we run
 * @param delta how many milliseconds until we run the method
 * @param deltaRepeat if this is a periodic, the time between
 *        the runs, otherwise 0.
 **/
void addCronJob(CronJob method,
		unsigned int delta,
		unsigned int deltaRepeat,
		void * data) 
{
  DeltaEntry * entry;
  DeltaEntry * pos;
  int last;
  int current;

  #if DEBUG_CRON
    LOG(LOG_CRON, "CRON: Adding job %x-%x to fire in %d CU\n", (int)method, (int)data, delta);
  #endif

  MUTEX_LOCK(&deltaListLock_);
  if (firstFree_ == -1) 
    {
      /* need to grow */
      DeltaEntry * tmp;
      int i;

      tmp = xmalloc(sizeof(DeltaEntry) * deltaListSize_ * 2, "addCronJob: grow deltaList");
      memcpy(tmp, deltaList_, deltaListSize_*sizeof(DeltaEntry));
      xfree(deltaList_, "addCronJob: old deltaList");
      deltaList_ = tmp;
      for (i=deltaListSize_;i<deltaListSize_*2;i++)
         deltaList_[i].next = i-1;
      deltaList_[deltaListSize_].next = -1;
      deltaListSize_ = deltaListSize_ * 2;
      firstFree_ = deltaListSize_-1;
    }
  entry = &deltaList_[firstFree_];
  entry->method = method;
  entry->data = data;
  entry->deltaRepeat = deltaRepeat;
  entry->delta = cronTime(NULL) + delta;
  if (firstUsed_ == -1) 
    {
      firstUsed_ = firstFree_;
      firstFree_ = entry->next;
      entry->next = -1; /* end of list */
      MUTEX_UNLOCK(&deltaListLock_);  
      /* interrupt sleeping cron-thread! */
      #if DEBUG_CRON
        LOG(LOG_CRON, "CRON: Sending signal to pid %d\n", cronPID);
      #endif
        if (cronPID != 0) kill(cronPID, SIGALRM);
      return;
    }
  /* no, there are jobs waiting */
  last = -1;
  current = firstUsed_;
  pos = &deltaList_[current];

  while (entry->delta > pos->delta) 
    {
      if (pos->next != -1) 
	    {
          last = current;
          current = pos->next;
          pos = &deltaList_[current];
        } 
	  else 
	    { /* append */
          pos->next = firstFree_;
          firstFree_ = entry->next;
          entry->next = -1;
          MUTEX_UNLOCK(&deltaListLock_); 
          return;
        }      
    }
  /* insert before pos */
  if (last == -1) 
    {
      firstUsed_ = firstFree_;
      if (cronPID != 0) kill(cronPID, SIGALRM);
    } 
  else 
    deltaList_[last].next = firstFree_;  
  firstFree_ = entry->next;
  entry->next = current;
  MUTEX_UNLOCK(&deltaListLock_); 
}

/**
 * Initialize this module.
 **/
void initKnownhosts() 
{
  int i;

  for (i=0;i<MAX_TEMP_HOSTS;i++)
     {
       tempHosts[i] = NULL;
	 }
  tempHostsNextSlot = 0;
  max_   = 128;
  count_ = 0;
  MUTEX_CREATE_RECURSIVE(&lock_);
  networkIdDirectory = getHostsDirectory();
  mkdirp(networkIdDirectory);
  hosts_ = xmalloc(sizeof(HostEntry) * max_, "initKnownhosts: hostlist");	      
  cronScanDirectoryDataHosts(NULL);
  addCronJob(&cronScanDirectoryDataHosts, 15 * cronMINUTES, 15 * cronMINUTES, NULL); 
}

/**
 * Remove all matching cron-jobs from the list.
 * @param method which method is listed?
 * @param repeat which repeat factor was chosen? 
 * @param data what was the data given to the method
 **/
void delCronJob(CronJob method,
		unsigned int repeat,
		void * data) 
{
  DeltaEntry * job;
  DeltaEntry * last;
  int jobId;

  LOG(LOG_CRON, "CRON: deleting job %x-%x\n", (int) method, (int) data);
  MUTEX_LOCK(&deltaListLock_);
  jobId = firstUsed_;
  if (jobId == -1) 
    {
	  MUTEX_UNLOCK(&deltaListLock_); 
      return; 
    }
  last = NULL;
  job = &deltaList_[jobId];
  while ( (job->method != method) || (job->data != data) || (job->deltaRepeat != repeat) ) 
    {
      last = job;
      if (job->next == -1) 
	    {
          MUTEX_UNLOCK(&deltaListLock_); 
         return; 
        }
      jobId = job->next;
      job = &deltaList_[jobId];    
    }
  if (last != NULL) last->next = job->next;
  else
    firstUsed_ = job->next;
  job->next = firstFree_;
  firstFree_ = jobId;
  MUTEX_UNLOCK(&deltaListLock_);
  /* ok, there may be more matches, go again! */
  delCronJob(method, repeat, data);
}

/**
 * Shutdown the knownhosts module.
 **/
void doneKnownhosts() 
{
  int i;

  /* delCronJob(&cronScanDirectoryDataHosts, 15 * cronMINUTES, NULL); */

  for (i=0;i<MAX_TEMP_HOSTS;i++)
     {
       freeNonNull(tempHosts[i], "HELO of host in temporary cache of knownhosts");
	 }
  MUTEX_DESTROY(&lock_);
  xfree(hosts_, "doneKnownhosts: hostlist");
  xfree(networkIdDirectory, "network id directory");
}

/**
 * This method must be called first! It reads
 * the config file and makes everything else
 * possible.
 **/
void initConfiguration() 
{
  MUTEX_CREATE(&configLock);
  /* setConfigurationString("FILES",
			 "gnunet.conf",
			 "~/.gnunet/gnunet.conf"); */
}

/* don't do anything, we use SIGALRM to abort
   the nanosleep */
static void sigalrmHandler(int sig) 
{
  #ifdef CRON_DEBUG
    LOG(LOG_CRON, "CRON: RECEIVED SIGALRM\n");
  #endif
}

/**
 * Initialize the cron service.
 **/
void initCron() 
{
  int i;
  struct sigaction sig;

  deltaListSize_ = INIT_CRON_JOBS;
  deltaList_ = xmalloc(sizeof(DeltaEntry) * deltaListSize_, "initCron: deltalist");
  for (i=0;i<deltaListSize_;i++)
    deltaList_[i].next = i-1;
  firstFree_ = deltaListSize_-1;
  MUTEX_CREATE(&deltaListLock_);
  firstUsed_  = -1;
  /* SA_NODEFER == SA_NOMASK but is available on linux+ */
  sig.sa_flags = SA_NODEFER;
  sig.sa_sigaction = NULL;
  sig.sa_handler = &sigalrmHandler;
  sigaction(SIGALRM, &sig, NULL);
}

/**
 * Catch sigpipe on send, and do nothing
 **/
static void catcher(int sig) 
{
  LOG(LOG_MESSAGE, "MESSAGE: signal %d caught (ignored)\n", sig);
  signal(sig, catcher);
}

/**
 * Initialize TCPIO module.
 * Installs a signal handler for SIGPIPE because read/write
 * operations on sockets can send that signal and we would
 * like to just ignore it.
 **/
void initTCPIO() 
{
  if (SIG_ERR == signal(SIGPIPE, SIG_IGN)) 
    {
      /* retry: install 'method'  as signal handler */
      if (SIG_ERR == signal(SIGPIPE, catcher))
        LOG(LOG_WARNING, "WARNING: could not install handler for SIGPIPE!\nAttempting to continue anyway.");
    }
}

/**
 * Initialize the statistics module.
 **/
void initStatistics() {
  statCounters = 0;
  cronTime(&startTime);
  MUTEX_CREATE(&statLock);
}

/**
 * Initialize the logging module.
 **/
void initLogging() 
{
  MUTEX_CREATE(&logMutex);
  resetLogging();
}

/**
 * Get a handle to a statistical entity.
 *
 * @param name a description of the entity
 * @return a handle for updating the associated value
 **/
int statHandle(char * name) 
{
  int i;
  unsigned long long * tmp;
  char ** tmp2;
  
  if (name == NULL) errexit("statHandle called with name being NULL\n");
  MUTEX_LOCK(&statLock);
  for (i=0;i<statCounters;i++)
    if (0 == strcmp(descriptions[i], name)) 
	  {
        MUTEX_UNLOCK(&statLock);
        return i;
      }
  
  tmp = xmalloc(sizeof(unsigned long long) * (statCounters+1), "statHandle: values (grows)");
  tmp2 = xmalloc(sizeof(char *) * (statCounters+1), "statHandle: descriptions (grows)");

  if (statCounters > 0) 
    { /* only grow if statcounters != 0 */ 
	  memcpy(tmp, values, sizeof(unsigned long long) * statCounters);
      memcpy(tmp2, descriptions, sizeof(char*) * statCounters);
      xfree(descriptions, "statHandle: old descriptions");
      xfree(values, "statHandle: old values");
    }
  
  values = tmp;
  values[statCounters] = 0;
  descriptions = tmp2;
  descriptions[statCounters] = STRDUP(name);
  statCounters++;
  MUTEX_UNLOCK(&statLock);   
  return statCounters-1;
}

static void getSizeRec(char * filename,
		       char * dirname,
		       unsigned long long * size) 
{
  struct stat buf;  
  char * fn;

  if (filename == NULL) return;
  if (dirname != NULL) 
    {
      fn = xmalloc(strlen(filename) + strlen(dirname) + 2, "allocating space for filename");
      fn[0] = '\0';
      if (strlen(dirname) > 0) 
	    {
          strcat(fn, dirname);      
          if (dirname[strlen(dirname)-1] != '/') strcat(fn, "/"); /* add tailing / if needed */
        } 
	  else
        strcat(fn, "/"); 
      if (filename[0] == '/') /* if filename starts with a "/", don't copy it */ strcat(fn, &filename[1]);
      else
        strcat(fn, filename);
    } 
  else
    fn = STRDUP(filename);

  if (0 != stat(fn, &buf)) 
    {
      LOG(LOG_DEBUG, "DEBUG: Can not stat %s (%s)\n", fn, strerror(errno));
      xfree(fn, "freeing filename");
      return;
    }
  *size += buf.st_size;
  if ( (S_ISDIR(buf.st_mode)) && (!S_ISLNK(buf.st_mode)) ) 
    {
      scanDirectory(fn, (DirectoryEntryCallback)&getSizeRec, size);
    }  
  xfree(fn, "freeing filename");
}

/**
 * Get the size of the file (or directory)
 * of the given file (in bytes).
 **/
unsigned long long getFileSize(char * filename) 
{
  unsigned long long size;

  size = 0;
  LOG(LOG_EVERYTHING, "EVERYTHING: getting file size of %s\n", filename);
  getSizeRec(filename, "", &size);
  return size;
}

/**
 * Manipulate statistics. Sets the statistics associated with the
 * handle to value.
 *
 * @param handle the handle for the value to change
 * @param value to what the value should be set
 **/
void statSet(const int handle,
	     const unsigned long long value) 
{
  MUTEX_LOCK(&statLock);
  if ( (handle < 0) || (handle >= statCounters) ) 
    {
      LOG(LOG_WARNING, "WARNING: invalid call to statSet, h=%d, statC=%d!\n", handle, statCounters);
      MUTEX_UNLOCK(&statLock);
      return;
    }
  values[handle] = value;
  MUTEX_UNLOCK(&statLock);
}

static void spaceJob() 
{
  int i;
  long long unitsFreed;
  long long oldNeed;
  long long needToFree;
 
  needToFree = spaceLeft - MIN_FREE;
  oldNeed = needToFree;
  for (i=0;i<dfc_count;i++) 
    driveFullCallbacks[i](&needToFree);
  unitsFreed = needToFree - oldNeed;
 
  spaceUsed -= unitsFreed;
  spaceLeft += unitsFreed;

  statSet(stat_handle_kb_used, (spaceUsed > 0 ? spaceUsed : 0));
  statSet(stat_handle_kb_remaining, (spaceLeft > 0 ? spaceLeft : 0));

  for (i=0;i<dfc_count+1;i++)
    semaphore_up(signalSemaphore);
  MUTEX_LOCK(&statusMutex);
  free_semaphore(signalSemaphore);
  signalSemaphore = NULL;
  MUTEX_UNLOCK(&statusMutex);
}

static void cronSpaceMonitor() 
{
  unsigned long long use;
  int i;

  LOG(LOG_CRON, "CRON: cronSpaceMonitor\n");
  MUTEX_LOCK(&statusMutex);
  use = 0;
  /* FIXME: deletion from gdb/tdb does not necessary affect getFileSize()
   * at all. After hitting the limit for the first time, we will use spaceJob() 
   * each time cronSpaceMonitor() gets called, before the db is reorganized 
   * (very costly). This might call for redesign. 
   */
  for (i=0;i<tracked_file_count;i++)
    use += getFileSize(tracked_files[i]);
  spaceUsed = use / 1024;
  spaceLeft /* KB */ = (diskQuota /* MB */ * 1024) - (use /* bytes */ / 1024);
  statSet(stat_handle_kb_used, (spaceUsed > 0 ? spaceUsed : 0));
  statSet(stat_handle_kb_remaining,(spaceLeft > 0 ? spaceLeft : 0));

  if (spaceLeft < MIN_FREE) 
    {
      if (signalSemaphore == NULL) 
	    {
          pthread_t job;

          signalSemaphore = new_semaphore(0);
          PTHREAD_CREATE(&job, (void * (*)()) &spaceJob, NULL, 1024*64);
          pthread_detach(job);
        }
    }
  MUTEX_UNLOCK(&statusMutex);
  LOG(LOG_CRON, "CRON: cronSpaceMonitor done.\n");
}

/**
 * Re-read the configuration for statuscalls.
 **/
 
void resetStatusCalls() 
{
  char * interfaces;
  char * ifcs;
  int start;

  MUTEX_LOCK(&statusMutex);
  /* interfaces = getConfigurationString("LOAD", "INTERFACES"); */
  
  interfaces = STRDUP(INTERFACES);
  ifcs = interfaces;
  /* fail if config-file is incomplete */
  if (interfaces == NULL) errexit("No network interfaces defined!\n");

  /* The string containing the interfaces is formatted in the following way: 
   * each comma is replaced by '\0' and the pointers to the beginning of every 
   * interface are stored
   */
  numInterfaces = 0;
  start = YES;
  while (1) 
    {
      if (*interfaces == '\0') 
	    {
          if (start == NO) numInterfaces++;      
          break;
        }
      if ( ((*interfaces>='a') && (*interfaces<='z')) || ((*interfaces>='A') && (*interfaces<='Z')) || ((*interfaces>='0') && (*interfaces<='9')) ) 
	    {
		  start = NO;         
        } 
	  else 
	    {
          if (*interfaces != ',') errexit("interfaces string (%s) invalid\n", ifcs);
          if (start == NO) 
		    {
	          start = YES;
	          numInterfaces++;
            }
        }
      interfaces++;
    }
  if (numInterfaces <= 0) errexit("ERROR: No network interfaces specified in he configuration file\n");

  interfacePtrs = xmalloc(sizeof(char*) * numInterfaces, "allocating space for interfaces");
  last_net_results = xmalloc(sizeof(NetworkStats) * numInterfaces, "allocating space for interface statistics");
  memset(last_net_results, 0, sizeof(NetworkStats) * numInterfaces);

  /* 2nd pass, this time remember the positions */
  interfaces = ifcs;
  numInterfaces = 0;
  start = YES;
  while (1) 
    {
      if (*interfaces=='\0') 
	    {
          if (start == NO) numInterfaces++;      
          break;
        }
      if ( ((*interfaces>='a') && (*interfaces<='z')) || ((*interfaces>='A') && (*interfaces<='Z')) || ((*interfaces>='0') && (*interfaces<='9')) ) 
	    {
          if (start == YES) 
		    {
	          start = NO;   
	          interfacePtrs[numInterfaces] = interfaces;	
            }
        } 
	  else 
	    {
          if (start == NO) 
		    {
	          start = YES;
	          *interfaces = '\0';	
	          numInterfaces++;
            }
        }
      interfaces++;
    }

  /* useBasicMethod = testConfigurationString("LOAD", "BASICLIMITING", "YES"); */

  useBasicMethod = BASICLIMITING;

  /* maxNetDownBPS = getConfigurationInt("LOAD", "MAXNETDOWNBPSTOTAL"); */

  maxNetDownBPS = MAXNETDOWNBPSTOTAL;

  if (maxNetDownBPS == 0) maxNetDownBPS = 50000;

  /* maxNetUpBPS = getConfigurationInt("LOAD", "MAXNETUPBPSTOTAL"); */

  maxNetUpBPS = MAXNETUPBPSTOTAL;

  if (maxNetUpBPS == 0) maxNetUpBPS = 50000;

  /* maxNetLoad = getConfigurationInt("LOAD", "MAXNETLOAD"); */
  
  maxNetLoad = MAXNETLOAD;

  if (maxNetLoad == 0 || useBasicMethod == YES) maxNetLoad = 100;

  /* maxCPULoad = getConfigurationInt("LOAD", "MAXCPULOAD"); */

  maxCPULoad = MAXCPULOAD;

  if (maxCPULoad == 0) maxCPULoad = 100;

  /* diskQuota = getConfigurationInt("LOAD", "DISKQUOTA"); */

  diskQuota = DISKQUOTA;

  if (diskQuota == 0) diskQuota = 1;

  MUTEX_UNLOCK(&statusMutex);  
  cronSpaceMonitor();
}

/**
 * The basic usage meter considers only gnunetd traffic.
 **/
static int networkUsageBasic() 
{
  cron_t now, elapsedTime;
  int upUsage;
  int downUsage;
  double upCurrent;
  double downCurrent;
  double currentLoad;
  
  MUTEX_LOCK(&statusMutex);
  /* If called less than 1 seconds ago, calc average 
     between avg. traffic of last second and the additional
     gained traffic. If more than 1 second between calls,
     set the avg bytes/sec as the usage of last second. */
  cronTime(&now);  
  elapsedTime = now - lastnettime;
  if (elapsedTime < cronSECONDS) 
    {
      upUsage = (int)((double)lastNetResultUp + (double)globalTrafficBetweenProc.last_out) / (1.0 + ((double)elapsedTime / (double)cronSECONDS));
      downUsage = (int)((double)lastNetResultDown + (double)globalTrafficBetweenProc.last_in) / (1.0 + ((double)elapsedTime / (double)cronSECONDS));
    } 
  else 
    {
      upUsage = (int)((double)globalTrafficBetweenProc.last_out / ( (double)elapsedTime / (double)cronSECONDS ));
      downUsage = (int)((double)globalTrafficBetweenProc.last_in / ( (double)elapsedTime / (double)cronSECONDS ));
      lastNetResultUp = upUsage;
      lastNetResultDown = downUsage;
      globalTrafficBetweenProc.last_in = 0;
      globalTrafficBetweenProc.last_out = 0;
      lastnettime = now;
    }
   
  upCurrent = (double)upUsage / (double)maxNetUpBPS;
  downCurrent = (double)downUsage / (double)maxNetDownBPS;

  if(upCurrent > downCurrent)
    currentLoad = upCurrent;
  else
    currentLoad = downCurrent;
 
/*
  fprintf(stderr, "DEBUG: incrB gti=%llu gto=%llu ",
        globalTrafficBetweenProc.last_in, globalTrafficBetweenProc.last_out);
  fprintf(stderr, " e=%d d=%d u=%d load=%e ret=%d\n", 
	(int)(elapsedTime),
    	downUsage, upUsage, currentLoad, (int)(100.0 * currentLoad));
*/

  MUTEX_UNLOCK(&statusMutex);
  return (int)(100.0 * currentLoad);
}

/**
 * The advanced usage meter takes into account all network traffic.
 * This might be problematic on systems where the same interface 
 * can have different capabilities for different types of traffic 
 * (like support for very fast local traffic but capable of 
 * handling only small-scale inet traffic).
 **/
static int networkUsageAdvanced() 
{
  int retval = -1;
  char line[MAX_PROC_LINE];
  unsigned long long rxnew, txnew;
  unsigned long long rxdiff, txdiff;
  int i=0;
  int ifnum;
  cron_t now, elapsedtime;
  
  #ifdef LINUX
    FILE * proc_net_dev;
    char * data;
  #else
    FILE * command;
  #endif

  MUTEX_LOCK(&statusMutex);
  /* If we checked /proc less than 2 seconds ago, don't do
     it again, but add internal gnunet traffic increments */
  cronTime(&now);  
  elapsedtime = now - lastnettime;
  if (elapsedtime < 2 * cronSECONDS) 
    {
      int upDelta;
      int downDelta;
      int newTempMax;

      if (elapsedtime < 1*cronSECONDS)   /* reqd. for delta calc. below */ elapsedtime = 1*cronSECONDS;

      upDelta = (100 * globalTrafficBetweenProc.last_out * cronSECONDS) / (elapsedtime * maxNetUpBPS);
      downDelta = (100 * globalTrafficBetweenProc.last_in * cronSECONDS) / (elapsedtime * maxNetDownBPS);

      /* be a bit conservative, think the load measured from /proc
       as current base load and add internal gnunet load to it.
       any better ideas? perhaps we should always average the load
       of past 2 secs or something... */
      if (lastNetResultUp + upDelta > lastNetResultDown + downDelta) newTempMax = lastNetResultUp + upDelta;
      else
        newTempMax = lastNetResultDown + downDelta;
         /*
         fprintf(stderr, "DEBUG: incrB gti=%llu gto=%llu ",
         globalTrafficBetweenProc.last_in, globalTrafficBetweenProc.last_out);
         fprintf(stderr, " e=%d d=%d u=%d newMax=%d\n", 
	    (int)(elapsedtime+1),
    	downDelta, upDelta, newTempMax);
        */

      MUTEX_UNLOCK(&statusMutex);
      return newTempMax;
    } 
  else 
    {
      globalTrafficBetweenProc.last_in = 0;
      globalTrafficBetweenProc.last_out = 0;
    }
  lastnettime = now;

  /* ok, full program... */
  if ( (maxNetUpBPS == 0) || (maxNetDownBPS == 0) ) 
    {
      lastNetResultUp = -1;
      lastNetResultDown = -1;
      MUTEX_UNLOCK(&statusMutex);
      return -1;
    }
  rxdiff = 0;
  txdiff = 0;
  
  #ifdef LINUX
    proc_net_dev = fopen(PROC_NET_DEV, "r");
    /* Try to open the file*/
    if (NULL == proc_net_dev) 
	  {
        LOG(LOG_ERROR, "ERROR: Could not open %s (%s)\n", PROC_NET_DEV, strerror(errno));
        MUTEX_UNLOCK(&statusMutex);
        lastNetResultUp = -1;
        lastNetResultDown = -1;
        return -1;
      }
    ifnum = 0;
    /* Parse the line matching the interface ('eth0') */
    while ( (!feof(proc_net_dev)) && ( ifnum < numInterfaces) ) 
	  {    
        fgets(line, MAX_PROC_LINE, proc_net_dev);
    
        for (i=0;i<numInterfaces;i++) 
	       {
             if (NULL != strstr(line, interfacePtrs[i]) ) 
		 	  {
	             data = (char*)strchr(line, ':');
	             data++;	
	             if (sscanf(data, "%llu %*s %*s %*s %*s %*s %*s %*s %llu", &rxnew, &txnew) != 2) 
			 	   {
	                 fclose(proc_net_dev);
	                 errexit("ERROR: reading interface data from %s\n", PROC_NET_DEV);
	               } 	
	             rxdiff += rxnew - last_net_results[ifnum].last_in;	  
	             txdiff += txnew - last_net_results[ifnum].last_out;
	             /*LOG(LOG_NOTHING,
	             "NOTHING: (%d<%d) %s -- rxdiff %llu = %llu - %llu; txdiff %llu = %llu - %llu\n",
	             ifnum, numInterfaces,
	             line,
	             rxdiff, rxnew, last_net_results[ifnum].last_in,
	             txdiff, txnew, last_net_results[ifnum].last_out);*/
	             last_net_results[ifnum].last_in = rxnew;
	             last_net_results[ifnum].last_out = txnew;
	             ifnum++;
	             break;
               }
           }
      }
    fclose(proc_net_dev);
  #else
    if (1) 
	  {
        MUTEX_UNLOCK(&statusMutex);
        return 0;
      }
    if ( ( command = popen("netstat -n -f inet -i", "r") ) == NULL ) 
	  {
        LOG(LOG_ERROR, "ERROR: could not open netstat\n");
        lastNetResultUp = -1;
        lastNetResultDown = -1;
        MUTEX_UNLOCK(&statusMutex);
        return -1;
      }
    ifnum = 0;
    while ( (!feof(command)) && (ifnum < numInterfaces ) ) 
	  {
        fgets(line, MAX_PROC_LINE, command);
        for (i=0; i < numInterfaces; i++) 
		   {
             if ( NULL != strstr(line, interfacePtrs[i]) ) 
			   {
	             if (sscanf(line, "%*s %*s %*s %*s %llu %*s %llu %*s %*s", &rxnew, &txnew) != 2 ) 
				   {
	                 pclose(command);
	                 errexit("ERROR: reading interface data using netstat\n");
	               }
	             rxdiff += rxnew - last_net_results[ifnum].last_in;
	             last_net_results[ifnum].last_in = rxnew;
	             txdiff += txnew - last_net_results[ifnum].last_out;
	             last_net_results[ifnum].last_out = txnew;
	             ifnum++;
	             break;
               } /* if match */
           }  /* for all interfaces, find match */
      } /* while: for all lines in proc */
    pclose(command);
  #endif
  
  lastNetResultUp = (100 * txdiff * cronSECONDS) / (elapsedtime * maxNetUpBPS);
  lastNetResultDown = (100 * rxdiff * cronSECONDS) / (elapsedtime * maxNetDownBPS);
	
  if (lastNetResultUp > lastNetResultDown) retval = lastNetResultUp;
  else
    retval = lastNetResultDown;  

  MUTEX_UNLOCK(&statusMutex);
  return(retval);  
}

/**
 * The following routine returns the percentage of available used
 * bandwidth.  Example: If 81 is returned this means that 81% of the
 * network bandwidth of the host is consumed.  The method
 * initStatusCalls() should be called before this routine is invoked.
 * If there is an error the method returns -1.
 **/
int networkUsage() 
{
  if ( useBasicMethod == YES )
    return networkUsageBasic();
  else
    return networkUsageAdvanced();
}

/**
 * The following routine returns a number between 0-100 (can be larger than 100
 * if the load is > 1) which indicates the percentage CPU usage.  
 * 
 * Before its first invocation the method initStatusCalls() must be called.
 * If there is an error the method returns -1
 **/
int cpuUsage()
{
  cron_t now, elapsedtime;
  static cron_t lastcputime  =  0; /* In seconds */
  static int lastcpuresult = -1;
  #ifndef HAVE_GETLOADAVG
  #ifdef LINUX
    int ret = -1;
    static int last_cpu_results[4] = { 0, 0, 0, 0 };
    FILE* proc_stat;
    char line[128];
    int user_read, system_read, nice_read, idle_read;
    int user, system, nice, idle;
    int usage_time=0, total_time=1;
  #endif
  #else
    double loadavg;
  #endif

  MUTEX_LOCK(&statusMutex);
  cronTime(&now);  
  elapsedtime = now - lastcputime;
  if (elapsedtime < 2 * cronSECONDS) 
    {
      MUTEX_UNLOCK(&statusMutex);
      return lastcpuresult;
    }
  lastcputime = now;
  
  #ifndef HAVE_GETLOADAVG
  #ifdef LINUX
    /* Get the current time in microseconds */

    proc_stat = fopen("/proc/stat", "r");
    /* Try to open the file*/
    if (NULL == proc_stat) 
	  {
        LOG(LOG_ERROR, "ERROR: Could not open /proc/stat");
        lastcpuresult = -1;
        MUTEX_UNLOCK(&statusMutex);
        return -1;
      }
    /* Get the first line with the data */
    if (fgets(line, 128, proc_stat)==NULL) 
	  {
        LOG(LOG_ERROR, "ERROR: reading /proc/stat file");
        fclose(proc_stat);
        lastcpuresult = -1;
        MUTEX_UNLOCK(&statusMutex);
        return -1;
      }
    if (sscanf(line, "%*s %i %i %i %i", &user_read, &system_read, &nice_read, &idle_read) != 4) 
	  {
        fclose(proc_stat);
        LOG(LOG_ERROR, "ERROR: decoding /proc/stat file");
        lastcpuresult = -1;
        MUTEX_UNLOCK(&statusMutex);
        return -1;
      }

    /* Store the current usage*/
    user = user_read - last_cpu_results[0];
    system = system_read - last_cpu_results[1];
    nice = nice_read - last_cpu_results[2];
    idle = idle_read - last_cpu_results[3];
  
    /* Calculate the % usage */
    if ((user + system + nice + idle) > 0) 
	  {
		usage_time = user + system + nice;
        total_time = usage_time + idle;
      }
    ret = (100 * usage_time) / total_time;

    /* Store the values for the next calculation*/
    last_cpu_results[0] = user_read;
    last_cpu_results[1] = system_read;
    last_cpu_results[2] = nice_read;
    last_cpu_results[3] = idle_read;
    fclose(proc_stat);

    lastcpuresult = ret;
    MUTEX_UNLOCK(&statusMutex);
    return ret;
  #else
    /* loadaverage not defined and not linux: error */
    lastcpuresult = -1;
    MUTEX_UNLOCK(&statusMutex);
    return -1;
  #endif
  #else
    if ( getloadavg(&loadavg, 1) == -1 ) 
	  {
        LOG(LOG_ERROR, "ERROR: getting load average\n");
        lastcpuresult = -1;
        MUTEX_UNLOCK(&statusMutex);
        return -1;
      }
    lastcpuresult = (int) (100 * loadavg);
    MUTEX_UNLOCK(&statusMutex);
    return lastcpuresult;
  #endif
}


/**
 * Get the load of the CPU relative to what is allowed.
 * @return the CPU load as a percentage of allowed
 *        (100 is equivalent to full load)
 **/
int getCPULoad() 
{
  static int lastRet = 0;
  int ret = (100 * cpuUsage()) / maxCPULoad;
  if (ret > lastRet) 
    lastRet = ret; /* quickly rise to higher loads */
  else 
    {
      ret = (ret + 7 * lastRet)/8;
      lastRet = ret; /* but only slowly go back! */
    }
  statSet(stat_handle_cpu_load, ret);
  return ret;
}

/**
 * Get the load of the network relative to what is allowed.
 * @return the network load as a percentage of allowed
 *        (100 is equivalent to full load)
 **/
int getNetworkLoad() 
{
  static int lastRet = 0;
  int ret = (100 * networkUsage()) / maxNetLoad;
  if (ret > lastRet) 
    lastRet = ret; /* quickly rise to higher loads */
  else 
    {
      ret = (ret + 7 * lastRet)/8;
      lastRet = ret; /* but only slowly go back! */
    }
  statSet(stat_handle_network_load, ret);
  return ret;
}

static void cronLoadUpdate(void * unused) 
{
  getCPULoad();
  getNetworkLoad();
}

/**
 * The following method is called in order to initialize the status calls
 * routines. After that it is safe to call each of the status calls separately
 * @return OK on success and SYSERR on error (or calls errexit).
 **/
void initStatusCalls() 
{
  LOG(LOG_DEBUG, "DEBUG: in initStatusCalls\n");
  MUTEX_CREATE_RECURSIVE(&statusMutex);
  tracked_file_count = 0;
  last_net_results = NULL; /* has numInterfaces entries */
  spaceLeft = 0; /* in KB */
  spaceUsed = 0; /* in KB */
  dfc_count = 0;
  interfacePtrs = NULL;
  last_net_results = NULL;
  tracked_files = NULL;
  driveFullCallbacks = NULL;
  globalTrafficBetweenProc.last_in = 0;
  globalTrafficBetweenProc.last_out = 0;
  stat_handle_kb_used = statHandle("# kb of storage used");
  stat_handle_kb_remaining = statHandle("# kb of storage remaining");
  stat_handle_network_load = statHandle("% of allowed network load");
  stat_handle_cpu_load = statHandle("% of allowed cpu load");
  cronTime(&lastnettime);  
  resetStatusCalls();
  networkUsage(); 
  cpuUsage();  
  /* every hour, check space consumption */
  addCronJob(&cronSpaceMonitor, 10 * cronMINUTES, 10 * cronMINUTES, NULL);
  addCronJob(&cronLoadUpdate, 10 * cronSECONDS, 10 * cronSECONDS, NULL);
}

/**
 * Obtain the identity information for the current node
 * (connection information), conInfo.
 * @return SYSERR on failure, OK on success
 **/
static int getAddressFromHostname(IPaddr * identity) 
{
  char * hostname;
  struct hostent * ip;

  hostname = getenv("HOSTNAME");  
  if (hostname == NULL) 
    {
      LOG(LOG_ERROR, "ERROR: Environment variable HOSTNAME not set!\n");
      return SYSERR;
    }
  LOG(LOG_DEBUG, "DEBUG: looking up $HOSTNAME (%s) to obtain local IP\n", hostname);
  ip = gethostbyname(hostname);
  if (ip == NULL) 
    {
      LOG(LOG_ERROR, "ERROR: Couldn't find IP for host '%s' (h_errno=%d)", hostname, h_errno);
      return SYSERR;
    }
  if (ip->h_addrtype != AF_INET) 
    {
      LOG(LOG_ERROR, "ERROR: getAddress: h_addrtype is not AF_INET (%d)!?", ip->h_addrtype);
      return SYSERR;
    }
  memcpy(identity, &((struct in_addr*)ip->h_addr_list[0])->s_addr, sizeof(struct in_addr));
  return OK;
}

static int getAddressFromIOCTL(IPaddr * identity) 
{
  struct ifreq ifr[MAX_INTERFACES];
  struct ifconf ifc;
  int sockfd,ifCount;
  int i;
  char * interfaces;

  interfaces = STRDUP(INTERFACES);
  /* interfaces = getConfigurationString("NETWORK", "INTERFACE"); */
  if (interfaces == NULL) 
    {
      LOG(LOG_ERROR, "ERROR: No interface specified in section NETWORK under INTERFACE!\n");
      return SYSERR; /* that won't work! */
    }

  sockfd = socket(PF_INET, SOCK_DGRAM, 0);
  if (sockfd == -1) return SYSERR;
  ifc.ifc_len = sizeof(ifr);
  ifc.ifc_buf = (char*)&ifr;
  
  if (ioctl(sockfd, SIOCGIFCONF, &ifc) == -1) 
    {
      LOG(LOG_WARNING, "WARNING: Could not obtain IP with ioctl\n");
      close(sockfd);
      return SYSERR;
    }
  ifCount = ifc.ifc_len / sizeof(struct ifreq);
  
  /* first, try to find exatly matching interface */
  for (i=0;i<ifCount;i++)
     {
       if (ioctl(sockfd, SIOCGIFADDR, &ifr[i]) != 0) continue;
       if (ioctl(sockfd, SIOCGIFFLAGS, &ifr[i]) != 0) continue;
       if (!(ifr[i].ifr_flags & IFF_UP)) continue; 
       if (strcmp((char*) interfaces, (char*) ifr[i].ifr_name) != 0) continue;
       memcpy(identity, &(((struct sockaddr_in *)&ifr[i].ifr_addr)->sin_addr), sizeof(struct in_addr));
       close(sockfd);
       return OK;
     }
  LOG(LOG_WARNING, "WARNING: Could not find interface %s in IOCTL, trying to find another one.\n", interfaces);
  /* if no such interface exists, take any interface but loopback */
  for (i=0;i<ifCount;i++)
     {
       if (ioctl(sockfd, SIOCGIFADDR, &ifr[i]) != 0) continue;
       if (ioctl(sockfd, SIOCGIFFLAGS, &ifr[i]) != 0) continue;
       if (!(ifr[i].ifr_flags & IFF_UP)) continue;
       if (strncmp("lo", (char*) ifr[i].ifr_name, 2) == 0) continue;
       memcpy(identity, &(((struct sockaddr_in *)&ifr[i].ifr_addr)->sin_addr), sizeof(struct in_addr));
       close(sockfd);
       return OK;
     }
  close(sockfd);
  LOG(LOG_WARNING, "WARNING: Could not obtain IP for interface %s using ioctl.\n", interfaces);
  return SYSERR;
}

/**
 * Get the IP address for the local machine.
 * @return SYSERR on error, OK on success
 **/
static int getAddress(IPaddr  * address){
  char * ipString;
  int retval;
  struct hostent * ip; /* for the lookup of the IP in gnunet.conf */

  retval = SYSERR;
  ipString = STRDUP(IP_ADDRESS);
  /* ipString = getConfigurationString("NETWORK", "IP"); */
  if (ipString == NULL) 
    {
      if (OK == getAddressFromIOCTL(address))
        retval = OK;
      else 
        retval = getAddressFromHostname(address);
    } 
  else 
	{
      LOG(LOG_DEBUG, "DEBUG: obtaining local IP address from hostname %s\n", ipString);
      ip = gethostbyname(ipString);
      if (ip == NULL) 
	    {
          LOG(LOG_ERROR, "ERROR: Couldn't resolve '%s' (%s)", ipString, hstrerror(h_errno));
          retval = SYSERR;
        } 
	  else 
	    {
          if (ip->h_addrtype != AF_INET) 
		    {
	          LOG(LOG_ERROR, "ERROR: getAddress: h_addrtype is not AF_INET (%d)!?", ip->h_addrtype);
	          retval = SYSERR;
            } 
		  else 
		    {
	          memcpy (address, &((struct in_addr*) ip->h_addr_list[0])->s_addr, sizeof(struct in_addr));
	          retval = OK;
            }
        }
      xfree(ipString, "ipString from configuration.getIP (strdup)");
    }
  return retval;
}

static void cronRefreshAddress(void * unused) 
{
  LOG(LOG_CRON, "CRON: enter cronRefreshAddress\n");
  if (SYSERR == getAddress(&myAddress)) LOG(LOG_FAILURE, "FAILURE: Could not determine IP address of the local machine!\n");
  LOG(LOG_CRON, "CRON: exit cronRefreshAddress\n");
}

/** 
 * Initialize identity module. Requires configuration.
 **/
void initAddress() 
{
  if (SYSERR == getAddress(&myAddress)) errexit("FATAL: Could not find IP for this host. Please provide the IP in the configuration file.\n");  
  addCronJob(&cronRefreshAddress, 2 * cronMINUTES, 2 * cronMINUTES, NULL);
}

/**
 * Initialize the util library. Use argc, argv and the given parser
 * for processing command-line options <bf>after</bf> the
 * configuration module was initialized, but <bf>before</bf> logging
 * and other configuration-dependent features are started.
 **/
int initUtil(int argc,
	     char * argv[]) 
{
  initRAND();
  initConfiguration();
  initStatistics();
  initCron();
  initTCPIO();  
  /* if (SYSERR == parser(argc, argv))return SYSERR; */
  /* readConfiguration(); */
  initLogging();
  initStatusCalls();
  initAddress(); 
  return OK;
}

/**
 * Shutdown TCPIO module.
 **/
void doneTCPIO() {
  /* should we restore the signal handler here? */
}

/**
 * Shutdown the status calls module.
 **/
void doneStatusCalls() 
{
  delCronJob(&cronSpaceMonitor, 1 * cronHOURS, NULL);
  delCronJob(&cronLoadUpdate, 10 * cronSECONDS, NULL);
  xfree(interfacePtrs[0], "doneStatusCalls: interface-string(s)");
  xfree(interfacePtrs, "doneStatusCalls: interfacePtrs array");
  if (last_net_results != NULL) xfree(last_net_results, "doneStatusCalls: network statistics");    
  freeNonNull(tracked_files, "doneStatusCalls: freeing list of tracked files");
  if (driveFullCallbacks != NULL) xfree(driveFullCallbacks, "doneStatusCalls: freeing list of out of space callbacks");
  MUTEX_DESTROY(&statusMutex);
}

/** 
 * Shutdown identity module.
 **/
void doneAddress() 
{
  delCronJob(&cronRefreshAddress, 2 * cronMINUTES, NULL);
}

/**
 * Make sure to call stopCron before calling this method!
 **/
void doneCron() 
{
  int i;

  i = firstUsed_;
  while (i != -1) 
    {
      freeNonNull(deltaList_[i].data, "cron job not executed but cron module unloaded, trying to reclaim memory");
      i = deltaList_[i].next;
    }
  MUTEX_DESTROY(&deltaListLock_);
  xfree(deltaList_, "doneCron");
  deltaList_ = NULL;
}

/**
 * Shutdown the statistics module.
 **/
void doneStatistics() 
{
  int i;

  MUTEX_DESTROY(&statLock);
  for (i=0;i<statCounters;i++)
    xfree(descriptions[i], "doneStatistics, description strings");
  freeNonNull(descriptions, "doneStatistics: descriptions");
  freeNonNull(values, "doneStatistics: values");
  descriptions = NULL;
  values = NULL;
}

void doneUtil() 
{
  doneTCPIO();
  doneStatusCalls();
  doneAddress();
  doneCron();
  /* doneConfiguration(); */
  doneStatistics();
  LOG(LOG_MESSAGE, "Shutdown complete.\n");
  doneLogging();  
}

/********************************************************
* initialize tables                                     *
*********************************************************/
void my_init_table(SQLHDBC hdbc, SQLHSTMT hstmt)
{
  SQLRETURN   rc;

  printf("\nmy_init_table:\n");

    /* drop table 'my_demo_param' if it already exists */
    printf(" creating table 'my_demo_param'\n");

    rc = SQLExecDirect(hstmt,"DROP TABLE if exists my_demo_param",SQL_NTS);
    mystmt(hstmt,rc);

    /* commit the transaction */
    rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT); 
    mycon(hdbc,rc);

    /* create the table 'my_demo_param' */
    rc = SQLExecDirect(hstmt,"CREATE TABLE my_demo_param(\
                              id   int,\
                              auto int primary key auto_increment,\
                              name varchar(20),\
                              timestamp timestamp(14))",SQL_NTS);
    mystmt(hstmt,rc);

    /* commit the transaction*/
    rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT);
    mycon(hdbc,rc);    
}

/********************************************************
* prints the statement resultset                        *
*********************************************************/
int my_print_resultset(SQLHSTMT hstmt)
{
  SQLRETURN   rc;
  SQLUINTEGER nRowCount=0, pcColDef;
  SQLCHAR     szColName[MAX_NAME_LEN];
  SQLCHAR     szData[MAX_COLUMNS][MAX_ROW_DATA_LEN]={""};
  SQLSMALLINT nIndex,ncol,pfSqlType, pcbScale, pfNullable;

    /* get total number of columns from the resultset */
    rc = SQLNumResultCols(hstmt,&ncol);
    mystmt(hstmt,rc);

    /* print the column names  and do the row bind */
    for(nIndex = 1; nIndex <= ncol; nIndex++)
    {
      rc = SQLDescribeCol(hstmt,nIndex,szColName, MAX_NAME_LEN+1, NULL,
                          &pfSqlType,&pcColDef,&pcbScale,&pfNullable);
      mystmt(hstmt,rc);

      printf(" %s\t",szColName);

      rc = SQLBindCol(hstmt,nIndex, SQL_C_CHAR, szData[nIndex-1],
                      MAX_ROW_DATA_LEN+1,NULL);
      mystmt(hstmt,rc);
    }

    printf("\n -------------------------------------------\n");

    /* now fetch row by row */
    rc = SQLFetch(hstmt);
    while(rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO)
    {
       nRowCount++;
       for(nIndex=0; nIndex< ncol; nIndex++)
         printf(" %s\t",szData[nIndex]);
     
       printf("\n");
       rc = SQLFetch(hstmt);
     }
     SQLFreeStmt(hstmt,SQL_UNBIND);

     printf("\n total rows fetched:%d\n", ((int) nRowCount));

     /* free the statement row bind resources */
     rc = SQLFreeStmt(hstmt, SQL_UNBIND);
     mystmt(hstmt,rc);

     /* free the statement cursor */
     rc = SQLFreeStmt(hstmt, SQL_CLOSE);
     mystmt(hstmt,rc);   
     
     return(nRowCount);
}

/********************************************************
* insert data using parameters                          *
*********************************************************/
void my_param_insert(SQLHDBC hdbc, SQLHSTMT hstmt)
{
  SQLRETURN   rc;
  SQLINTEGER  id;
  SQLCHAR     name[50];

  printf("\nmy_param_insert:\n");

    /* prepare the insert statement with parameters */
    rc = SQLPrepare(hstmt,"INSERT INTO my_demo_param(id,name) VALUES(?,?)",SQL_NTS);
    mystmt(hstmt,rc);

    /* now supply data to parameter 1 and 2 */
    rc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, 
                          SQL_C_LONG, SQL_INTEGER, 0,0,
                          &id, 0, NULL);
    mystmt(hstmt,rc);

    rc = SQLBindParameter(hstmt, 2, SQL_PARAM_INPUT, 
                          SQL_C_CHAR, SQL_CHAR, 0,0,
                          name, sizeof(name), NULL);
    mystmt(hstmt,rc);

    /* now insert 10 rows of data */
    for (id = 0; id < 10; id++)
    {
      sprintf(name,"MySQL%d",((int) id));

      rc = SQLExecute(hstmt);
      mystmt(hstmt,rc);
    }
    
    /* Free statement param resorces */
    rc = SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
    mystmt(hstmt,rc);
    
    /* Free statement cursor resorces */
    rc = SQLFreeStmt(hstmt, SQL_CLOSE);
    mystmt(hstmt,rc);

    /* commit the transaction */
    rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT); 
    mycon(hdbc,rc);

    /* Now fetch and verify the data */
    rc = SQLExecDirect(hstmt, "SELECT * FROM my_demo_param",SQL_NTS);
    mystmt(hstmt,rc);

    printf("My_print_resultset=%d!\n", my_print_resultset(hstmt));

    /* assert(10 == my_print_resultset(hstmt));*/
}

/********************************************************
* update data using parameters                          *
*********************************************************/
void my_param_update(SQLHDBC hdbc, SQLHSTMT hstmt)
{
  SQLRETURN   rc;
  SQLINTEGER  id=9, nRowCount;
  SQLCHAR     name[]="update";

  printf("\nmy_param_update:\n");

    /* prepare the insert statement with parameters */
    rc = SQLPrepare(hstmt,"UPDATE my_demo_param set name = ? WHERE id = ?",SQL_NTS);
    mystmt(hstmt,rc);

    /* now supply data to parameter 1 and 2 */    
    rc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, 
                          SQL_C_CHAR, SQL_CHAR, 0,0,
                          name, sizeof(name), NULL);
    mystmt(hstmt,rc);

    rc = SQLBindParameter(hstmt, 2, SQL_PARAM_INPUT, 
                          SQL_C_LONG, SQL_INTEGER, 0,0,
                          &id, 0, NULL);
    mystmt(hstmt,rc);

    /* now execute the update statement */
    rc = SQLExecute(hstmt);
    mystmt(hstmt,rc);
    
    /* check the rows affected by the update statement */
    rc = SQLRowCount(hstmt, &nRowCount);
    mystmt(hstmt,rc);
    printf("\n total rows updated:%d\n", ((int) nRowCount));
    /* assert( nRowCount == 1);*/
    
    /* Free statement param resorces */
    rc = SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
    mystmt(hstmt,rc);
    
    /* Free statement cursor resorces */
    rc = SQLFreeStmt(hstmt, SQL_CLOSE);
    mystmt(hstmt,rc);

    /* commit the transaction */
    rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT); 
    mycon(hdbc,rc);

    /* Now fetch and verify the data */
    rc = SQLExecDirect(hstmt, "SELECT * FROM my_demo_param",SQL_NTS);
    mystmt(hstmt,rc);


    printf("My_print_resultset=%d!\n", my_print_resultset(hstmt));
    /* assert(10 == my_print_resultset(hstmt));*/
}

/********************************************************
* delete data using parameters                          *
*********************************************************/
void my_param_delete(SQLHDBC hdbc, SQLHSTMT hstmt)
{
  SQLRETURN   rc;
  SQLINTEGER  id, nRowCount;

  printf("\nmy_param_delete:\n");
    
    /* supply data to parameter 1 */    
    rc = SQLBindParameter(hstmt, 1, SQL_PARAM_INPUT, 
                          SQL_C_LONG, SQL_INTEGER, 0,0,
                          &id, 0, NULL);
    mystmt(hstmt,rc);

    /* execute the DELETE STATEMENT to delete 5th row  */
    id = 5;
    rc = SQLExecDirect(hstmt,"DELETE FROM my_demo_param WHERE id = ?",SQL_NTS);
    mystmt(hstmt,rc);
    
    /* check the rows affected by the update statement */
    rc = SQLRowCount(hstmt, &nRowCount);
    mystmt(hstmt,rc);
    printf(" total rows deleted:%d\n", ((int) nRowCount));
    /* assert( nRowCount == 1);*/

    /* execute the DELETE STATEMENT to delete 8th row  */
    id = 8;
    rc = SQLExecDirect(hstmt,"DELETE FROM my_demo_param WHERE id = ?",SQL_NTS);
    mystmt(hstmt,rc);
    
    /* check the rows affected by the update statement */
    rc = SQLRowCount(hstmt, &nRowCount);
    mystmt(hstmt,rc);
    printf(" total rows deleted:%d\n", ((int) nRowCount));
    /* assert( nRowCount == 1); */
    
    /* Free statement param resorces */
    rc = SQLFreeStmt(hstmt, SQL_RESET_PARAMS);
    mystmt(hstmt,rc);
    
    /* Free statement cursor resorces */
    rc = SQLFreeStmt(hstmt, SQL_CLOSE);
    mystmt(hstmt,rc);

    /* commit the transaction */
    rc = SQLEndTran(SQL_HANDLE_DBC, hdbc, SQL_COMMIT); 
    mycon(hdbc,rc);

    /* Now fetch and verify the data */
    rc = SQLExecDirect(hstmt, "SELECT * FROM my_demo_param",SQL_NTS);
    mystmt(hstmt,rc);

    /* assert(8 == my_print_resultset(hstmt)); */

     printf("My_print_resultset=%d!\n", my_print_resultset(hstmt));
}
/********************************************************
* main routine                                          *
*********************************************************/
int main(int argc, char *argv[])
{  
    /*
     *  show the usage string when the user asks for this
    */    
      printf("***********************************************\n");
      printf("usage: my_param [DSN] [UID] [PWD] \n");  
      printf("***********************************************\n");     

    /*
     * if connection string supplied through arguments, overrite
     * the default one..
    */
    for(narg = 1; narg < argc; narg++)
    {     
      if ( narg == 1 )
        mydsn = argv[1];
      else if ( narg == 2 )
        myuid = argv[2];
      else if ( narg == 3 )
        mypwd = argv[3];      
    }   

    /* 
     * connect to MySQL server
    */
    myconnect(&henv,&hdbc,&hstmt); 

    /* 
     * initialize table
    */
    /* my_init_table(hdbc, hstmt); */

    /* 
     * insert data using parameters
    */

    /* printf("pass 01\n"); */
    if (OK != initUtil(argc, argv))return SYSERR;
    /* printf("pass 02\n"); */
    initKeyService();
    /* printf("pass 03\n"); */
    /* initTransports();   */
    /* printf("pass 04\n"); */
    initKnownhosts();
    /* printf("pass 05 number of hosts=%d!\n", count_); */

    /* printf("count for forEachHost=%d!\n", forEachHost(&printHostInfo, 0, NULL)); 

    printf("pass 06a\n"); */
    printf("forEachHost2=%d!\n", forEachHost2());
    /* printf("pass 06b\n"); */
    doneKnownhosts();
    /* printf("pass 07\n"); */
  
    /* printf("pass 08\n"); */
    doneKeyService();
    /* printf("pass 09\n"); */
    doneUtil();
    /* printf("pass 10\n"); */
  
    /* my_param_insert(hdbc, hstmt); */

    /* 
     * parameter update 
    */
    /* my_param_update(hdbc, hstmt); */

    /* 
     * parameter delete
    */
    /* my_param_delete(hdbc, hstmt); */

    /* 
     * disconnect from the server, by freeing all resources
    */
    mydisconnect(&henv,&hdbc,&hstmt);

  return(0);
} 



